#ifndef BOOTSTRAP_CUDADEVICE_H
#define BOOTSTRAP_CUDADEVICE_H


#include <vector>
#include <vector_types.h>
#include <curand_kernel.h>
#include "utils.h"
#include "bootstrap.h"
#include "devicemanager.h"
#include "graph/graph.h"
#include "operation.h"
#include <list>


struct CudaTimers {
    Timer initialize;

    Timer generateSeedsCpu;
    Timer loadIterationCount1s;
    Timer sortSeedsCpu;
    Timer copySeedsCpuToGpu;
    Timer copySeedsAndIterationCountsCpuToGpu;

    Timer initPrng;
    Timer bootstrapWarpFriendly;
    Timer bootstrapNonWarpFriendly;

    Timer sortSeedsGpu;
    Timer generateSeedsGpu;
    Timer copySeedsGpuToCpu;
    Timer loadIterationCount2s;
    Timer copyIterationCountsCpuToGpu;

    Timer copyResultFromGpuToCpu;

    Timer finalize;

    Timer getOptimalExecutionPath;
};

struct CudaResult : Result {
    int device_id;

    struct CudaTimers bootstrapCudaTimers;
};


//struct CudaBootstrapState {
//    int *seedsDev = nullptr;
//    int *iterCountDev = nullptr;
//    float *inputDev = nullptr;
//    float *outputDev = nullptr;
//
//    std::vector<int> seedsHost;
//    std::vector<int> iterCountHost;
//    std::vector<float> inputHost;
//    std::vector<float> outputHost;
//};


class CudaDevice : public Device {
public:
    CudaDevice(int device_id, std::vector<int> *iter_counts);

    virtual ~CudaDevice();

    void saveEstimators();

    void setData(std::vector<float> cumulative_input,
                 int n_iterations,
                 float p,
                 Flags flags);

    int getMaxNIterations();

    void initialize();

    void generateSeedsCpu();
    void loadIterationCount1();
    void sortSeedsCpu();
    void copySeedsCpuToGpu();
    void copySeedsAndIterationCountsCpuToGpu();

    void initPrng();
    void bootstrapWarpFriendly();
    void bootstrapNonWarpFriendly();

    void sortSeedsGpu();
    void generateSeedsGpu();
    void copySeedsGpuToCpu();
    void loadIterationCount2();
    void copyIterationCountsCpuToGpu();
    void copyResultFromGpuToCpu();

    void finalize();

    std::unique_ptr<CudaResult> bootstrapGetResult();

    virtual std::string getDeviceName() const {
        return std::string(device_properties.name);
    }

    virtual std::string getDeviceType() const {
        return std::string("GPU");
    }

    std::list<Operation> findOptimalExecutionPath(int n_iterations, bool print_path=false);

    double getCost(int n_iterations) override;
    double getDCost(int n_iterations) override;

    double executionPathCost(std::list<Operation> &executionPath, int n_iterations);
    double executionPathDCost(std::list<Operation> &executionPath, int n_iterations);

    std::vector<int> getIterationCounts(int seeds_from, int seeds_to);

    std::vector<int> *all_iter_counts;

private:
    Graph<Operation> operation_graph;

    int device_id;
    cudaDeviceProp device_properties;

    dim3 grid_dim;
    dim3 block_dim;

    float p;
    int n_iterations;
    Flags flags;

    struct CudaTimers timers;
    Timer iterations_time;

    void setCudaTopology();

    std::mt19937 randomGenerator;

    void initializeRandomGenerator();

    std::vector<int> seeds_host;
    std::vector<int> iteration_counts_host;
    std::vector<float> bootstrap_input_host;
    std::vector<float> bootstrap_output_host;

    int *seeds_dev;
    int *iteration_counts_dev;
    float *bootstrap_input_dev;
    float *bootstrap_output_dev;
    curandState_t *prng_states_dev;

    int *iterations_output_dev;
    std::vector<int> iterations_output_host;


    void buildOperationGraph();
    void bootstrap();
    void loadIterationCounts();
};

#endif //BOOTSTRAP_CUDADEVICE_H
