#ifndef BOOTSTRAP_HISTOGRAM_H
#define BOOTSTRAP_HISTOGRAM_H

#include <vector>
#include <limits>
#include <iostream>
#include "utils.h"
#include <json/json.h>


template <typename Input_T, typename Hist_T>
class Histogram {
public:
    Histogram(std::vector<Input_T> input)
            : input(input) {

        minVal = *std::min_element(input.cbegin(), input.cend());
        maxVal = *std::max_element(input.cbegin(), input.cend());

        reorganize();
        normalize();
    }

    void reorganize();
    void reorganize(int n_buckets, Input_T from_bucket, Input_T to_bucket);
    void normalize();
    std::pair<Input_T, Input_T> getBucketInterval(int bucket_num) const;
    const Hist_T& operator[](int bucket_num) const;

    Input_T getBucketsStart() const {return bucket_start;}
    Input_T getBucketsEnd() const {return bucket_start + bucket_size * (buckets.size() - 2);}

    int getNumBuckets() const {return (int) buckets.size() - 2;}

    template <typename T, typename U>
    friend nlohmann::json& operator<<(nlohmann::json &json, const Histogram<T, U> &histogram);

    Input_T minVal, maxVal;

private:
    std::pair<std::pair<Input_T, Input_T>, int> getAutoBucketSize();

    std::vector<Input_T> input;

    Input_T bucket_start;
    Input_T bucket_size;
    std::vector<Hist_T> buckets;
};

template <typename Input_T, typename Hist_T>
void Histogram<Input_T, Hist_T>::reorganize(int n_buckets, Input_T from_bucket, Input_T to_bucket) {
    buckets = std::vector<Hist_T>(n_buckets + 2, (Hist_T) 0);

    bucket_size = (to_bucket - from_bucket) / n_buckets;
    bucket_start = from_bucket;

    for (auto f : input) {
        int bucket_num = (f - bucket_start) / bucket_size + 1;
        bucket_num = std::min(bucket_num, (int) buckets.size() - 1);
        bucket_num = std::max(bucket_num, 0);
        buckets[bucket_num] += 1;
    }
}

template <typename Input_T, typename Hist_T>
std::pair<Input_T, Input_T> Histogram<Input_T, Hist_T>::getBucketInterval(int bucket_num) const {
    Input_T
            start = bucket_start + bucket_num * bucket_size,
            end = bucket_start + (bucket_num + 1) * bucket_size;

    float end_f = (float) end;
    end_f -= 0.0001;
    end = (Input_T) end_f;

    int adjusted_bucket_num = bucket_num+1;
    if (adjusted_bucket_num < 0 || adjusted_bucket_num >= (int) buckets.size()) {
        throw Exception("Histogram index must be in the interval [-1, n_buckets]!");
    }

    if (adjusted_bucket_num == 0) {
        start = std::numeric_limits<Input_T>::min();
    }
    if (adjusted_bucket_num == (int) buckets.size() - 1) {
        end = std::numeric_limits<Input_T>::max();
    }
    return std::make_pair(start, end);
}

template <typename Input_T, typename Hist_T>
const Hist_T& Histogram<Input_T, Hist_T>::operator[](int bucket_num) const {
    int adjusted_bucket_num = bucket_num + 1;
    if (adjusted_bucket_num < 0 || adjusted_bucket_num >= (int) buckets.size()) {
        throw Exception("Histogram index must be in the interval [-1, n_buckets]!");
    }

    return buckets[adjusted_bucket_num];
}

template <typename Input_T, typename Hist_T>
std::pair<std::pair<Input_T, Input_T>, int> Histogram<Input_T, Hist_T>::getAutoBucketSize() {
    Input_T
            bucket_start = 0,
            bucket_size = 10;

    int n_buckets = 10;

    std::cout << "Min: " << minVal << std::endl;
    std::cout << "Max: " << maxVal << std::endl;

    bucket_start = minVal;
    bucket_size = (maxVal - minVal) / n_buckets;

    if (bucket_size == 0) {
        bucket_size = 1;
    }

    return std::make_pair(std::make_pair(bucket_start, bucket_size), n_buckets);
}

template <typename Input_T, typename Hist_T>
void Histogram<Input_T, Hist_T>::normalize() {
    Hist_T sum = 0;
    for (auto f : buckets) {
        sum += f;
    }
    double factor = 1.0 / sum;

    for (Hist_T &f : buckets) {
        f *= factor;
    }
}

template <typename Input_T, typename Hist_T>
void Histogram<Input_T, Hist_T>::reorganize() {
    auto temp = getAutoBucketSize();
    Input_T bucket_start = temp.first.first;
    Input_T bucket_size = temp.first.second;
    int n_buckets = temp.second;

    reorganize(n_buckets, bucket_start, bucket_start + bucket_size * n_buckets);
}

template <typename Input_T, typename Hist_T>
nlohmann::json &operator<<(nlohmann::json &json, const Histogram<Input_T, Hist_T> &histogram) {
    for (int i = -1; i < (int) histogram.buckets.size() - 1; i++) {
        nlohmann::json bucket;
        auto bucket_interval = histogram.getBucketInterval(i);
        bucket["bucket_from"] = bucket_interval.first;
        bucket["bucket_to"] = bucket_interval.second;
        bucket["bucket_value"] = histogram[i];
        json.push_back(bucket);
    }
    return json;
}

#endif //BOOTSTRAP_HISTOGRAM_H
