#ifndef BOOTSTRAP_UTILS_H
#define BOOTSTRAP_UTILS_H

#include <chrono>
#include <utility>
#include <exception>
#include <string>
#include <vector>
#include <algorithm>
#include <fstream>
#include <driver_types.h>
#include <cuda_runtime_api.h>
#include <iostream>
#include <typeinfo>
#include <experimental/filesystem>
#include <json/json.h>


std::pair<int, int> getTwoFactors(int integer);

class Exception : std::exception {
public:
    Exception(std::string message) :
    _what(message) {}

    const char* what() const noexcept {
        return _what.c_str();
    }

    std::string _what;
};

class TimerException : Exception {
public:
    TimerException(std::string what) : Exception(what) {};
};

class Timer {
public:
    Timer();
    virtual ~Timer() {};

    double s() const;
    double ms() const;
    double us() const;
    double ns() const;

    void start();
    void stop();
    void reset();
    void restart();

private:
    typedef std::chrono::steady_clock my_clock_t;
    typedef my_clock_t::time_point time_point_t;
    typedef my_clock_t::duration duration_t;

    bool stopped;
    time_point_t start_time;
    time_point_t end_time;
    duration_t total_time;

    duration_t getTotalDuration() const;
    time_point_t now() const;

    friend Timer operator+(const Timer &, const Timer &);
    friend Timer operator-(const Timer &, const Timer &);
};

// Shamelessly copied from
// http://stackoverflow.com/questions/865668/how-to-parse-command-line-arguments-in-c
class InputParser {
public:
    InputParser (int &argc, char **argv){
        for (int i=1; i < argc; ++i) {
            auto option = std::string(argv[i]);

            // Support option=value
            size_t equals_pos = option.find("=");
            if (equals_pos != std::string::npos) {
                auto first = option.substr(0, equals_pos);
                auto second = option.substr(equals_pos+1);
                this->tokens.push_back(first);
                this->tokens.push_back(second);
            }
            else {
                this->tokens.push_back(std::string(argv[i]));
            }
        }
    }
    /// @author iain
    const std::string& getCmdOption(const std::string &option) const {
        std::vector<std::string>::const_iterator itr;
        itr =  std::find(this->tokens.begin(), this->tokens.end(), option);
        if (itr != this->tokens.end() && ++itr != this->tokens.end()){
            return *itr;
        }
        static const std::string empty_string("");
        return empty_string;
    }
    /// @author iain
    bool cmdOptionExists(const std::string &option) const{
        return std::find(this->tokens.begin(), this->tokens.end(), option)
               != this->tokens.end();
    }
private:
    std::vector <std::string> tokens;
};

bool file_exists(const std::string& filename);
bool file_exists(const std::experimental::filesystem::path& filename);


#define gpuErrchk(ans) { _gpuAssert((ans), __FILE__, __LINE__); }

void _gpuAssert(cudaError_t code, const char *file, int line, bool abort = true);

template <typename T>

void write_vector(std::string filename, const std::vector<T> &vec) {
    if (file_exists(filename)) {
        throw Exception(filename + " already exists!");
    }

    std::ofstream output(filename, std::ios::out | std::ios::binary | std::ios::ate);
    const char *start = (const char *) vec.data();
    output.write(start, vec.size() * sizeof (T));
    output.close();
}

template <typename T>
void read_vector(std::string filename, std::vector<T> &vec) {
    std::ifstream in(filename, std::ifstream::ate | std::ifstream::binary);
    long size_bytes = in.tellg();
    unsigned int bytes_per_element = sizeof (T);
    unsigned int n_elements = (unsigned int) (size_bytes / bytes_per_element);

    if (size_bytes % bytes_per_element != 0) {
        throw Exception(std::string("File does not contain an integer number of elements of type ") + std::string(typeid(T).name()));
    }

//    std::cout << "File has " << n_elements << " elements" << std::endl;
//    std::cout << "File is " << size_bytes<< " bytes" << std::endl;

    vec.resize(n_elements);

    in.seekg(std::ios_base::beg);
    in.read((char *) vec.data(), size_bytes);
}

template <typename T>
bool in_interval(T x, T interval_start, T interval_end) {
    return interval_start <= x && x <= interval_end;
}

// Returns true if a is within p*100 percent of b
template <typename T>
bool within_fraction(T a, T b, double p) {
    T b_max = (T) (b * (1.0 + p));
    T b_min = (T) (b * (1.0 - p));
    return in_interval(a, b_min, b_max);
}

extern nlohmann::json json;
extern int cuda_device;

#endif //BOOTSTRAP_UTILS_H
