#include "bootstrap.h"
#include <cpudevicemanager.h>
#include <device.h>
#include <cudadevicemanager.h>
#include <math_functions.h>


std::ostream& operator<<(std::ostream& out, const Flags& flags) {
    if (!flags) {
        out << "NONE";
    }
    else {
        int printed = 0;
#define ADD_BOOTSTRAP_FLAG_PRINTOUT(flag_name) \
        if (flags & Flags::flag_name) {\
            if (printed) {\
                out << " | ";\
            }\
            out << #flag_name;\
            printed = 1;\
        }
        ADD_BOOTSTRAP_FLAG_PRINTOUT(SORT_CPU)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(SORT_GPU)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(GROUP_INTO_WARPS)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(THRESHOLD)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(MANUAL_LOAD_BALANCE)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(GENERATE_ITER_CTS)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(GENERATE_SEEDS_CPU)
        ADD_BOOTSTRAP_FLAG_PRINTOUT(GENERATE_SEEDS_GPU)
    }
    return out;
}

Bootstrap::Bootstrap(std::string iter_counts_file)
    : iter_counts_file(iter_counts_file)
{
    device_managers.push_back(std::unique_ptr<DeviceManager>(new CpuDeviceManager()));
    device_managers.push_back(std::unique_ptr<DeviceManager>(new CudaDeviceManager(iter_counts_file)));
}

void Bootstrap::prepareInput(std::vector<float> input) {
    cumulativeInput = std::vector<float>(input.size() + 1);

    cumulativeInput[0] = 0.0f;
    for (int i = 0; i < (int) input.size(); ++i) {
        cumulativeInput[i+1] = cumulativeInput[i] + input[i];
    }
}

std::vector<float> Bootstrap::bootstrap(std::vector<float> input,
                                        int n_iterations, float p, Flags flags,
                                        float manual_gpu_fraction) {
    total_time.start();

    prepareInput(input);

    float gpu_fraction = manual_gpu_fraction;
    std::vector<int> device_iterations(2);

    std::vector<std::shared_ptr<Device>> devices;
    for (auto &device_manager : device_managers) {
        auto devices_this_manager = device_manager->getDevices();
        devices.insert(devices.end(), devices_this_manager.cbegin(), devices_this_manager.cend());
    }

    std::cout << "Detected devices:" << std::endl;
    for (auto &dev : devices) {
        std::cout << " - " << dev->getDeviceName() << " (" << dev->getDeviceType() << ")" << std::endl;
    }

    if (!(flags & Flags::MANUAL_LOAD_BALANCE)) {
        printf("Auto determining load balance\n");
        device_iterations = getLoadBalance(input, n_iterations, p);
    }
    else {
        if (!(0.0f <= gpu_fraction && gpu_fraction <= 1.0f)) {
            throw BootstrapException("gpu_fraction must be in the interval [0.0, 1.0]!\n");
        }

        int gpu_iterations = (int) (gpu_fraction * n_iterations);
        int cpu_iterations = n_iterations - gpu_iterations;
        device_iterations[0] = cpu_iterations;
        device_iterations[1] = gpu_iterations;
    }

    int total_iterations = 0;
    for (auto n_iter : device_iterations) {
        total_iterations += n_iter;
    }

    if (*std::min(device_iterations.cbegin(), device_iterations.cend()) < 0) {
        throw Exception("A device has been assigned less than 0 iterations!");
    }

    if (total_iterations != n_iterations) {
        throw Exception("Sum of iterations across devices does not equal initial number of iterations!");
    }

    std::cout << "Running " << n_iterations << " iterations of bootstrap on time series of length " << input.size() << std::endl;

    for (int dev_idx = 0; dev_idx < (int) device_managers.size(); dev_idx++) {
        if (device_iterations[dev_idx] > 0) {
            device_managers[dev_idx]->runAsync(cumulativeInput, device_iterations[dev_idx], p, flags, std::max(1, (int) std::thread::hardware_concurrency() - 2));
        }
    }

    auto total_avg = std::vector<float>();
    std::vector<std::unique_ptr<Result>> results;
    for (int dev_idx = 0; dev_idx < (int) device_managers.size(); dev_idx++) {
        if (device_iterations[dev_idx] > 0) {
            auto device_results = std::move(device_managers[dev_idx]->getResults());
            for (auto &device_result : device_results) {
                total_avg.insert(total_avg.end(), device_result->averages.cbegin(), device_result->averages.cend());
            }
        }
    }

    return total_avg;
}

std::vector<int> Bootstrap::getLoadBalance(std::vector<float> input, int n_iterations, float p) {
    Timer t;
    t.start();

    std::vector<int> res((unsigned int) 2);
//    res[0] = 0;
//    res[1] = n_iterations;
//    return res;

    double gpu_fraction = 0.5;

    int cpu_iterations = min(20000000, (int) (n_iterations * (1-gpu_fraction)));
    int gpu_iterations = n_iterations - cpu_iterations;

    double cpu_cost = device_managers[0]->getEstimate(cpu_iterations);
    double cpu_dCost = device_managers[0]->getDEstimate(cpu_iterations);
    double gpu_cost = device_managers[1]->getEstimate(gpu_iterations);
    double gpu_dCost = device_managers[1]->getDEstimate(gpu_iterations);

    nlohmann::json optimization_history, cpu_history, gpu_history;
    double converge_rate = 0.7;
    int i;
    for (i = 0; i < 100; i++, converge_rate *= 1.0) {
        if (cpu_cost < gpu_cost) {
            gpu_iterations -= (int) (converge_rate * (gpu_cost - cpu_cost) / (cpu_dCost + gpu_dCost));
            if (gpu_iterations < 0) {
                gpu_iterations = 0;
            }
            cpu_iterations = n_iterations - gpu_iterations;
        } else {
            cpu_iterations += (int) (converge_rate * (gpu_cost - cpu_cost) / (cpu_dCost + gpu_dCost));
            if (cpu_iterations < 0) {
                cpu_iterations = 0;
            }
            gpu_iterations = n_iterations - cpu_iterations;
        }

        cpu_cost = device_managers[0]->getEstimate(cpu_iterations);
        cpu_dCost = device_managers[0]->getDEstimate(max(1, cpu_iterations));
        gpu_cost = device_managers[1]->getEstimate(gpu_iterations);
        gpu_dCost = device_managers[1]->getDEstimate(max(1, gpu_iterations));

        cpu_history.push_back(cpu_cost);
        gpu_history.push_back(gpu_cost);

        std::cout << "============================================" << std::endl;
        std::cout << "CPU cost " << cpu_iterations << " iterations: " << cpu_cost << ", dCost: " << cpu_dCost << std::endl;
        std::cout << "GPU cost " << gpu_iterations << " iterations: " << gpu_cost << ", dCost: " << gpu_dCost << std::endl;

        if (within_fraction(cpu_cost, gpu_cost, 0.01)) {
            break;
        }
    }

    optimization_history["cpu"] = cpu_history;
    optimization_history["gpu"] = gpu_history;

    std::cout << "============================================" << std::endl;
    std::cout << "Optimization time: " << t.ms() << "ms" << std::endl;
    std::cout << "============================================" << std::endl;

    nlohmann::json optimization;
    optimization["iterations"] = i;
    optimization["optimization_history"] = optimization_history;
    optimization["optimization_time"] = t.ms();
    optimization["gpu_iterations"] = gpu_iterations;
    optimization["cpu_iterations"] = cpu_iterations;
    optimization["cpu_cost"] = cpu_cost;
    optimization["gpu_cost"] = gpu_cost;
    optimization["cpu_dCost"] = cpu_dCost;
    optimization["gpu_dCost"] = gpu_dCost;

    json["load_balancer"] = optimization;

    res[0] = cpu_iterations;
    res[1] = gpu_iterations;
    return res;
}

//nlohmann::json& operator<<(nlohmann::json& json_obj, const Bootstrap &bootstrap) {
//    using nlohmann::json;
//
//    json device_json;
//
//    json cpu_json, cpu_timers;
//    cpu_json["device_type"] = std::string("cpu");
//    cpu_json["device_name"] = std::string("AMD Ryzen 7 1800X Eight-Core Processor");
//    cpu_json["n_iterations"] = 123;
//    cpu_timers["total_ns"] = bootstrap.cpu_time.ns();
//
//    cpu_json["timers"] = cpu_timers;
//
//    std::cout << bootstrap.bootstrapCUDA->getDeviceName() << std::endl;
//    json gpu_json, gpu_timers;
//    gpu_json["device_type"] = std::string("gpu");
//    gpu_json["device_id"] = bootstrap.bootstrapCUDA->getDeviceId();
//    gpu_json["device_name"] = bootstrap.bootstrapCUDA->getDeviceName();
//    gpu_json["n_iterations"] = 123123;
//    gpu_timers["total_ns"] = bootstrap.gpu_time.ns();
//    gpu_timers["kernel_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_kernel_time.ns();
//    gpu_timers["memcpy_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_memcpy_time.ns();
//    gpu_timers["malloc_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_memory_management_time.ns();
//    gpu_timers["initprng_kernel_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_kernel_time.ns();
//    gpu_timers["initprng_memcpy_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_memcpy_time.ns();
//    gpu_timers["initprng_malloc_time_ns"] = bootstrap.bootstrapCUDA->bootstrap_memory_management_time.ns();
//
//    gpu_json["timers"] = gpu_timers;
//
//    device_json.push_back(cpu_json);
//    device_json.push_back(gpu_json);
//
//    json_obj["total_time_ns"] = bootstrap.total_time.ns();
//    json_obj["n_iterations"] = 123+123123;
//
//    json_obj["devices"] = device_json;
//
//    return json_obj;
//}
