#include <cpudevice.h>
#include <options.h>

std::string
CpuDevice::getDeviceName() const {
    std::ifstream cpuinfo("/proc/cpuinfo");
    std::string cur_line("");
    for ( ; cur_line.find("model name") == std::string::npos; std::getline(cpuinfo, cur_line)) {
        ;
    }
    size_t colon_pos = cur_line.find(":");
    size_t start_of_model_name = colon_pos + 2;
    return cur_line.substr(start_of_model_name);
}


std::vector<float>
CpuDevice::run(std::vector<float> cumulative_input,
               int n_iterations,
               float p,
               Flags flags) {
    int input_size = (int) cumulative_input.size() - 1;

    auto avgs = std::vector<float>((unsigned long) n_iterations);
    std::uniform_int_distribution<int> uniformIntDistribution(0, input_size - 1);
    std::geometric_distribution<int> geometricDistribution(p);

    for (int i = 0; i < n_iterations; i++) {

        float tot = 0;
        for (int available = input_size; available > 0; ) {
            // Draw interval start
            int start = uniformIntDistribution(randomGenerator);

            // Draw interval length
            int block_length = std::min(available, geometricDistribution(randomGenerator));

            // Calculate interval end
            int end = start + block_length;

            // Clamp at input size
            int end_clamped = std::min(end, input_size);

            // The portion that would be wrapped around
            int end_wrapped_around = std::max(0, end - input_size);

            // Interval is now [start, end_clamped) + [0, end_wrapped_around)
            available -= end_clamped - start + end_wrapped_around;

            // cumulativeInput[n] = sum(input[:n]) (Python notation)
            tot += cumulative_input[end_clamped] - cumulative_input[start] + cumulative_input[end_wrapped_around];
        }

        avgs[i] = tot / input_size;
    }
    return avgs;
}

CpuDevice::CpuDevice(int thread_id) : thread_id(thread_id)
{
    initializeRandomGenerator();

    using std::experimental::filesystem::path;
    path estimatorSaveDir = options.estimatorsDir / path(getDeviceName());
    std::experimental::filesystem::create_directories(estimatorSaveDir);
    estimator_ptr = std::shared_ptr<Estimator>(new Estimator(estimatorSaveDir / path("CPU bootstrap")));
    estimator_ptr->setAlpha(0.001);
}

void CpuDevice::initializeRandomGenerator() {
    unsigned seed = (unsigned int) std::chrono::system_clock::now().time_since_epoch().count();
    randomGenerator = std::mt19937(seed);
}

double CpuDevice::getCost(int n_iterations) {
    if (!estimator_ptr->canEstimate()) {
        return 0.0;
    }
    return estimator_ptr->estimate(n_iterations);
}

double CpuDevice::getDCost(int n_iterations) {
    if (!estimator_ptr->canEstimate()) {
        return 0.0;
    }
    return estimator_ptr->dEstimate(n_iterations);
}

void CpuDevice::train(int n_iterations, double cost) {
    estimator_ptr->train(n_iterations, cost);
}

CpuDevice::~CpuDevice() {
//    estimator_ptr->save();
}
