#include <cpudevicemanager.h>
#include <device.h>
#include <cpudevice.h>
#include <utils.h>


CpuDeviceManager::CpuDeviceManager()
{
    devices = std::vector<std::shared_ptr<Device>>();
    devices.push_back(std::shared_ptr<Device>(new CpuDevice(0)));
}

void CpuDeviceManager::runAsync(std::vector<float> cumulative_input, int n_iterations, float p, Flags flags, int max_threads) {
    n_threads = max_threads;
    future = std::move(
            std::async(std::launch::async, &CpuDeviceManager::runThreadedBootstrap, this, cumulative_input, n_iterations, p, flags)
    );
}

std::vector<std::unique_ptr<Result>> CpuDeviceManager::runThreadedBootstrap(std::vector<float> cumulative_input, int n_iterations, float p, Flags flags) {
    Timer total_time;
    total_time.start();

    nlohmann::json cpu;

    std::vector<std::shared_ptr<Device>> threads;
    threads.push_back(devices[0]);
    for (int i = 1; i < n_threads; i++) {
        threads.push_back(std::shared_ptr<CpuDevice>(new CpuDevice(i)));
    }

    std::vector<std::future<std::vector<float>>> cpu_future;
    int iterations_per_thread = n_iterations / n_threads;

    cpu["iterations_per_thread"] = iterations_per_thread;

    Timer bootstrap_timer;
    bootstrap_timer.start();
    // From i == 0 is handled by the main thread, after all the other threads have been launched. If n_threads == 1,
    // we are not spawning additional threads.
    for (int i = 1; i < n_threads; i++) {
        int from = i * iterations_per_thread;
        int to = (i+1) * iterations_per_thread;
        if (i+1 == n_threads) {
            to = n_iterations;
        }
        // Need to move (emplace does that) the std::future this returns, else this becomes synchronous
        // See "Notes" at: http://en.cppreference.com/w/cpp/thread/async
        cpu_future.emplace_back(std::async(std::launch::async, &CpuDevice::run, static_cast<CpuDevice *>(threads[i].get()), cumulative_input, to - from, p, flags));
    }

    // Have the main thread do its part of the work before gathering all the results
    auto main_thread_result = static_cast<CpuDevice *>(threads[0].get())->run(cumulative_input, iterations_per_thread, p, flags);
    std::cout << main_thread_result.size() << std::endl;

    auto result_vector = std::vector<std::unique_ptr<Result>>();

    std::string device_name;
    for (int i = 0; i < n_threads; i++) {
        std::vector<float> cur_avg;
        if (i == 0) {
            cur_avg = std::move(main_thread_result);
        }
        else {
            cur_avg = cpu_future[i-1].get();
        }

        std::unique_ptr<Result> res(new CpuResult());
        res->averages = cur_avg;
        res->device_name = threads[i]->getDeviceName();
        res->device_type = threads[i]->getDeviceType();
        res->stats.total_time = total_time;
        res->stats.p = p;
        res->stats.n_iterations = n_iterations;
        device_name = threads[i]->getDeviceName();

        result_vector.emplace_back(std::move(res));
    }

    bootstrap_timer.stop();
    std::cout << "CPU devices total time: " << bootstrap_timer.ms() << "ms" << std::endl;
    std::cout << "Training CPU model with " << n_iterations << " = " << bootstrap_timer.ms() << std::endl;
    static_cast<CpuDevice *>(devices[0].get())->train(n_iterations, bootstrap_timer.ms());
    static_cast<CpuDevice *>(devices[0].get())->estimator_ptr->save();
    total_time.stop();

    cpu["execution_time"] = total_time.ms();
    cpu["device_name"] = device_name;
    json["cpu"] = cpu;

    return result_vector;
}
