#include <cudadevice.h>
#include <kernels.h>
#include <queue>
#include <thrust/sort.h>
#include <thrust/device_ptr.h>
#include <thrust/execution_policy.h>
#include <options.h>

#define DEVICE_SYNCHRONIZE

CudaDevice::CudaDevice(int device_id, std::vector<int> *iter_counts)
        : device_id(device_id),
          all_iter_counts(iter_counts)
{
    cudaSetDevice(device_id);
    gpuErrchk( cudaDeviceSynchronize() );
    cudaFuncSetCacheConfig(_bootstrap_block_long, cudaFuncCachePreferL1);
    gpuErrchk( cudaDeviceSynchronize() );
    cudaGetDeviceProperties(&device_properties, device_id);
    gpuErrchk( cudaDeviceSynchronize() );
    initializeRandomGenerator();
    gpuErrchk( cudaDeviceSynchronize() );
    buildOperationGraph();
    gpuErrchk( cudaDeviceSynchronize() );
}

void CudaDevice::buildOperationGraph() {
    Operation source("Source");

    Operation initialize(getDeviceName(), "init", "Initialize", &CudaDevice::initialize, this);

    Operation generate_seeds_cpu(getDeviceName(), "gen_seeds_cpu", "Generate seeds CPU", &CudaDevice::generateSeedsCpu, this);

    Operation sort_seeds_cpu(getDeviceName(), "sort_seeds_cpu", "Sort seeds on the CPU", &CudaDevice::sortSeedsCpu, this);

    Operation load_iteration_counts1(getDeviceName(), "load_iter_1", "Get iteration counts 1", &CudaDevice::loadIterationCount1, this);

    Operation load_iteration_counts2(getDeviceName(), "load_iter_2", "Get iteration counts 2", &CudaDevice::loadIterationCount2, this);

    Operation copy_seeds_gpu_to_cpu(getDeviceName(), "cp_seeds_gpu_cpu", "Copy seeds from GPU to CPU", &CudaDevice::copySeedsGpuToCpu, this);

    Operation copy_seeds_cpu_to_gpu(getDeviceName(), "cp_seeds_cpu_gpu", "Copy seeds from CPU to GPU", &CudaDevice::copySeedsCpuToGpu, this);

    Operation copy_iteration_counts_cpu_to_gpu(getDeviceName(), "cp_iter_cpu_gpu", "Copy iteration counts from CPU to GPU", &CudaDevice::copyIterationCountsCpuToGpu, this);

    Operation copy_seeds_and_iteration_counts_cpu_to_gpu(getDeviceName(), "cp_seeds_iter_cpu_gpu", "Copy seeds and iteration counts from CPU to GPU", &CudaDevice::copySeedsAndIterationCountsCpuToGpu, this);

    Operation generate_seeds_gpu(getDeviceName(), "gen_seeds_gpu", "Generate seeds GPU", &CudaDevice::generateSeedsGpu, this);

    Operation sort_seeds_gpu(getDeviceName(), "sort_seeds_gpu", "Sort seeds on the GPU", &CudaDevice::sortSeedsGpu, this);

    Operation init_prng1(getDeviceName(), "init_prng_1", "Initialize PRNG 1", &CudaDevice::initPrng, this);

    Operation init_prng2(getDeviceName(), "init_prng_2", "Initialize PRNG 2", &CudaDevice::initPrng, this);

    Operation bootstrap_warp_friendly(getDeviceName(), "warp_friendly", "Warp-friendly bootstrap", &CudaDevice::bootstrapWarpFriendly, this);

    Operation bootstrap_non_warp_friendly(getDeviceName(), "non_warp_friendly", "Non warp-friendly bootstrap", &CudaDevice::bootstrapNonWarpFriendly, this);

    Operation copy_result_gpu_to_cpu(getDeviceName(), "cp_result_gpu_cpu", "Copy result from GPU to CPU", &CudaDevice::copyResultFromGpuToCpu, this);

    Operation finalize(getDeviceName(), "fin", "Finalize", &CudaDevice::finalize, this);

    Operation sink("Sink");

#define ADD_NODE(node_name)\
    operation_graph.add_node(node_name.id, node_name);

    ADD_NODE(source)
    ADD_NODE(initialize)
    ADD_NODE(generate_seeds_cpu)
    ADD_NODE(load_iteration_counts1)
    ADD_NODE(sort_seeds_cpu)
    ADD_NODE(copy_seeds_cpu_to_gpu)
    ADD_NODE(init_prng1)
    ADD_NODE(init_prng2)
    ADD_NODE(bootstrap_warp_friendly)
    ADD_NODE(copy_seeds_and_iteration_counts_cpu_to_gpu)
    ADD_NODE(sort_seeds_gpu)
    ADD_NODE(generate_seeds_gpu)
    ADD_NODE(copy_seeds_gpu_to_cpu)
    ADD_NODE(load_iteration_counts2)
    ADD_NODE(copy_iteration_counts_cpu_to_gpu)
    ADD_NODE(bootstrap_non_warp_friendly)
    ADD_NODE(copy_result_gpu_to_cpu)
    ADD_NODE(finalize)
    ADD_NODE(sink)

#undef ADD_NODE

#define ADD_EDGE(FROM, TO)\
    operation_graph.add_edge(FROM.id, TO.id);

    ADD_EDGE(source, initialize)
    ADD_EDGE(initialize, generate_seeds_cpu)
    ADD_EDGE(generate_seeds_cpu, load_iteration_counts1)
    ADD_EDGE(load_iteration_counts1, sort_seeds_cpu)
    ADD_EDGE(sort_seeds_cpu, copy_seeds_cpu_to_gpu)
    ADD_EDGE(copy_seeds_cpu_to_gpu, init_prng1)
    ADD_EDGE(init_prng1, bootstrap_warp_friendly)
    ADD_EDGE(bootstrap_warp_friendly, copy_result_gpu_to_cpu)

    ADD_EDGE(load_iteration_counts1, copy_seeds_and_iteration_counts_cpu_to_gpu)
    ADD_EDGE(copy_seeds_and_iteration_counts_cpu_to_gpu, sort_seeds_gpu)
    ADD_EDGE(sort_seeds_gpu, init_prng1)

    ADD_EDGE(initialize, generate_seeds_gpu)
    ADD_EDGE(generate_seeds_gpu, copy_seeds_gpu_to_cpu)
    ADD_EDGE(copy_seeds_gpu_to_cpu, load_iteration_counts2)
    ADD_EDGE(load_iteration_counts2, copy_iteration_counts_cpu_to_gpu)
    ADD_EDGE(copy_iteration_counts_cpu_to_gpu, sort_seeds_gpu)

    ADD_EDGE(generate_seeds_gpu, init_prng2)
    ADD_EDGE(init_prng2, bootstrap_non_warp_friendly)
    ADD_EDGE(bootstrap_non_warp_friendly, copy_result_gpu_to_cpu)
    ADD_EDGE(copy_result_gpu_to_cpu, finalize)
    ADD_EDGE(finalize, sink)

#undef ADD_EDGE
}

void CudaDevice::setData(std::vector<float> cumulative_input, int n_iterations, float p, Flags flags) {
    cudaSetDevice(device_id);
//    cudaDeviceReset();

    bootstrap_input_host = cumulative_input;
    this->p = p;
    this->n_iterations = n_iterations;
    this->flags = flags;
    iteration_counts_dev = nullptr;

    setCudaTopology();
}

void CudaDevice::initializeRandomGenerator() {
    gpuErrchk( cudaDeviceSynchronize() );
    std::cout << "Setting device to " << device_id << std::endl;
    cudaSetDevice(device_id);
    gpuErrchk( cudaDeviceSynchronize() );

    unsigned seed = (unsigned int) std::chrono::system_clock::now().time_since_epoch().count();
    randomGenerator = std::mt19937(seed);
}

void CudaDevice::setCudaTopology() {
    cudaSetDevice(device_id);

    int threads_per_block = 1024;
    // Round up to nearest threads_per_block
    int n_blocks = n_iterations / threads_per_block;
    if (n_iterations % threads_per_block != 0) {
        n_blocks++;
    }
    auto grid_dim3 = getTwoFactors(n_blocks);
    int grid_dim_x = grid_dim3.first;
    int grid_dim_y = grid_dim3.second;

    int block_dim_x = threads_per_block;
    int block_dim_y = 1;

    grid_dim = dim3((unsigned int) grid_dim_x, (unsigned int) grid_dim_y, 1);
    block_dim = dim3((unsigned int) block_dim_x, (unsigned int) block_dim_y, 1);
}

CudaDevice::~CudaDevice() {
    cudaSetDevice(device_id);
    cudaDeviceReset();
}

std::unique_ptr<CudaResult> CudaDevice::bootstrapGetResult() {
    cudaSetDevice(device_id);

    std::unique_ptr<CudaResult> res(new CudaResult());
    res->averages = bootstrap_output_host;
    res->stats.n_iterations = n_iterations;
    res->stats.p = p;
    res->bootstrapCudaTimers = timers;
    res->device_id = device_id;
    res->device_name = getDeviceName();
    res->device_type = getDeviceType();

    return res;
}

std::list<Operation> CudaDevice::findOptimalExecutionPath(int n_iterations, bool print_path) {
    // Bellman-Ford
    timers.getOptimalExecutionPath.start();

    typedef double dist_type_t;
    typedef Graph<Operation> OperationGraph;

    std::unordered_map<OperationGraph::g_id, dist_type_t> dist;
    std::unordered_map<OperationGraph::g_id, OperationGraph::g_id> prev;

    std::vector<OperationGraph::g_id> queue;

    OperationGraph::g_id source_id = 0;
    for (auto node : operation_graph.nodes) {
        OperationGraph::g_id node_id = node.first;
        if (node.second->is_source_node()) {
            source_id = node_id;
            queue.emplace_back(node_id);
            dist[node_id] = 0;
        }
        else {
            queue.emplace_back(node_id);
            dist[node_id] = std::numeric_limits<dist_type_t>::max();
        }
        // Note: Not setting prev[node_id] on purpose
    }

    for (int i = 0; i < (int) operation_graph.nodes.size(); i++) {
        for (auto node_id : queue) {
            for (auto edge : operation_graph.nodes[node_id]->edges_out) {
                auto neighbour_ptr = edge->to;
                OperationGraph::g_id neighbour_id = neighbour_ptr->id;
                auto cost = neighbour_ptr->data.cost((int) bootstrap_input_host.size(), n_iterations, p);
                dist_type_t alt = dist[node_id] + cost;
                if (alt < dist[neighbour_id]) {
                    prev[neighbour_id] = node_id;
                    dist[neighbour_id] = alt;
                }
            }
        }
    }

    // Skipping check for negative cycles, this is a DAG

    std::list<Operation> executionPath;
    for (auto node : operation_graph.nodes) {
        if (node.second->is_sink_node()) {
            OperationGraph::g_id node_id = node.first;
            node_id = prev[node_id];
            while (node_id != source_id) {
                executionPath.push_front(operation_graph.nodes[node_id]->data);
                node_id = prev[node_id];
                if (node_id == prev[node_id]) {
                    break;
                }
            }
            break;
        }
    }

    if (print_path) {
        std::cout << "Optimal execution path:" << std::endl;
        for (auto it = executionPath.cbegin(); it != executionPath.cend(); it++) {
            std::cout << it->operationName << std::endl;
        }
    }

    timers.getOptimalExecutionPath.stop();
    return executionPath;
}

void CudaDevice::generateSeedsCpu() {
    timers.generateSeedsCpu.start();

    std::uniform_int_distribution<int> int_distribution(0, std::numeric_limits<int>::max());
    for (int i = 0; i < n_iterations; i++) {
        seeds_host[i] = int_distribution(randomGenerator);
    }

    timers.generateSeedsCpu.stop();
}

void CudaDevice::loadIterationCounts() {
    cudaSetDevice(device_id);

//    int prev = seeds_host[0];
//    bool sorted = true;
//    for (int i = 1; i < (int) seeds_host.size(); i++) {
//        if (seeds_host[i] < prev) {
//            sorted = false;
//            break;
//        }
//        prev = seeds_host[i];
//    }
//    if (sorted) {
//        std::cout << "========================== SORTED!!! ==========================" << std::endl;
//    } else {
//        std::cout << "========================== NOT SORTED!!! ==========================" << std::endl;
//    }

    for (int i = 0; i < (int) seeds_host.size(); i++) {
        iteration_counts_host[i] = (*all_iter_counts)[seeds_host[i]];
    }
}

void CudaDevice::sortSeedsCpu() {
    cudaSetDevice(device_id);

    timers.sortSeedsCpu.start();

    int max_val = *std::max_element(iteration_counts_host.begin(), iteration_counts_host.end());

    auto sorted = std::vector<std::vector<int>>((unsigned int) max_val+1);
    for (int i = 0; i < n_iterations; i++) {
        sorted[iteration_counts_host[i]].push_back(seeds_host[i]);
    }

    int idx = 0;
    seeds_host = std::vector<int>();
    for (int i = 0; i < max_val+1; i++) {
        for (auto seed : sorted[i]) {
            seeds_host.push_back(seed);
            idx++;
        }

//        timers.group_into_warps_time.start();
//        if (flags & Flags::GROUP_INTO_WARPS) {
//            for (; idx % 32 != 0; idx++) {
//                seeds.push_back(sorted[i][0]);
//            }
//        }
//        timers.group_into_warps_time.stop();
    }
    n_iterations = idx;

    timers.sortSeedsCpu.stop();
}

void CudaDevice::copySeedsCpuToGpu() {
    cudaSetDevice(device_id);

    timers.copySeedsCpuToGpu.start();
    if (n_iterations != (int) seeds_host.size()) {
        throw Exception("Number of seeds does not match the number of iterations!");
    }

    gpuErrchk( cudaMemcpy(seeds_dev, seeds_host.data(), sizeof (int) * n_iterations, cudaMemcpyHostToDevice) );

#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    timers.copySeedsCpuToGpu.stop();
}

void CudaDevice::bootstrap() {
    cudaSetDevice(device_id);

#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    Timer t;
    t.start();

    printf("Grid: [%d, %d, %d]\n", grid_dim.x, grid_dim.y, grid_dim.z);
    printf("Block: [%d, %d, %d]\n", block_dim.x, block_dim.y, block_dim.z);

    gpuErrchk( cudaMalloc((void **) &bootstrap_output_dev, sizeof (float) * n_iterations) );

    if (options.custom_variance) {
        _bootstrap_force_variance
        <<<grid_dim, block_dim, bootstrap_input_host.size() * sizeof (float)>>>
        (
            bootstrap_input_dev,
            bootstrap_input_host.size(),
            bootstrap_output_dev,
            n_iterations,
            prng_states_dev,
            p,
            options.mu,
            options.sigma
        );
    }
    else {
        if (bootstrap_input_host.size() > device_properties.sharedMemPerBlock / sizeof (float)) {
            std::cout << "Using _bootstrap_block_long" << std::endl;
            _bootstrap_block_long
            <<<grid_dim, block_dim>>>
            (
                bootstrap_input_dev,
                bootstrap_input_host.size(),
                bootstrap_output_dev,
                n_iterations,
                prng_states_dev,
                p
            );
        }
        else {
            std::cout << "Using _bootstrap_block" << std::endl;
            _bootstrap_block
            <<<grid_dim, block_dim, bootstrap_input_host.size() * sizeof (float)>>>
            (
                bootstrap_input_dev,
                bootstrap_input_host.size(),
                bootstrap_output_dev,
                n_iterations,
                prng_states_dev,
                p
            );
        }
    }


    // Sync so the kernel finishes
#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    t.stop();

    std::cout << "Bootstrap time: " << t.ms() << "ms" << std::endl;
    // Sync so the kernel finishes
#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    gpuErrchk( cudaPeekAtLastError() );
}

void CudaDevice::copySeedsAndIterationCountsCpuToGpu() {
    copySeedsCpuToGpu();
    copyIterationCountsCpuToGpu();
}

void CudaDevice::sortSeedsGpu() {
    cudaSetDevice(device_id);

    timers.sortSeedsGpu.start();
    // Sorts [iteration_counts_dev, iteration_counts_dev+n_iterations)
    thrust::sort_by_key(thrust::device, iteration_counts_dev, iteration_counts_dev+n_iterations, seeds_dev);

    gpuErrchk( cudaFree(iteration_counts_dev) );
    iteration_counts_dev = nullptr;

    timers.sortSeedsGpu.stop();
}

void CudaDevice::generateSeedsGpu() {
    cudaSetDevice(device_id);

    timers.generateSeedsGpu.start();

    _generate_seeds
    <<<grid_dim, block_dim>>>
    (
        (int) randomGenerator(), seeds_dev, n_iterations
    );

#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    thrust::sort(thrust::device, seeds_dev, seeds_dev+n_iterations);
    timers.generateSeedsGpu.stop();
}

void CudaDevice::copySeedsGpuToCpu() {
    cudaSetDevice(device_id);

    timers.copySeedsGpuToCpu.start();
    gpuErrchk( cudaMemcpy(seeds_host.data(), seeds_dev, sizeof (int) * n_iterations, cudaMemcpyDeviceToHost) );
    timers.copySeedsGpuToCpu.stop();
}

void CudaDevice::copyIterationCountsCpuToGpu() {
    cudaSetDevice(device_id);

    timers.copyIterationCountsCpuToGpu.start();
    gpuErrchk( cudaMalloc((void **) &iteration_counts_dev, sizeof (int) * n_iterations) );
    gpuErrchk( cudaMemcpy(iteration_counts_dev, iteration_counts_host.data(), sizeof (int) * n_iterations, cudaMemcpyHostToDevice) );
#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    timers.copyIterationCountsCpuToGpu.stop();
}

void CudaDevice::initialize() {
    cudaSetDevice(device_id);

    timers.initialize.start();

    seeds_host = std::vector<int>((unsigned int) n_iterations);
    iteration_counts_host = std::vector<int>((unsigned int) n_iterations);
    bootstrap_output_host = std::vector<float>((unsigned int) n_iterations);
    gpuErrchk( cudaMalloc((void **) &seeds_dev, sizeof (unsigned int) * n_iterations) );
    gpuErrchk( cudaMalloc((void **) &prng_states_dev, sizeof (curandState_t) * n_iterations) );
    gpuErrchk( cudaMalloc((void **) &bootstrap_input_dev,  sizeof (float) * bootstrap_input_host.size()) );
    gpuErrchk( cudaMemcpy(bootstrap_input_dev, bootstrap_input_host.data(), sizeof (float) * bootstrap_input_host.size(), cudaMemcpyHostToDevice) );

//#ifdef DEVICE_SYNCHRONIZE
//    gpuErrchk( cudaDeviceSynchronize() );
//#endif

    timers.initialize.stop();
}

void CudaDevice::finalize() {
    cudaSetDevice(device_id);

    timers.finalize.start();

#define FREE_AND_WARN(cuda_pointer)\
    if (cuda_pointer != nullptr) {\
        std::cout << #cuda_pointer " != nullptr, freeing..." << std::endl;\
        gpuErrchk( cudaFree(cuda_pointer) );\
        cuda_pointer = nullptr;\
    }

    FREE_AND_WARN(bootstrap_input_dev)
    FREE_AND_WARN(bootstrap_output_dev)
    FREE_AND_WARN(prng_states_dev)
    FREE_AND_WARN(seeds_dev)
    FREE_AND_WARN(iteration_counts_dev)

#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif

    timers.finalize.stop();
}

void CudaDevice::loadIterationCount1() {
    timers.loadIterationCount1s.start();
    loadIterationCounts();
    timers.loadIterationCount1s.stop();
}

void CudaDevice::loadIterationCount2() {
    timers.loadIterationCount2s.start();
    loadIterationCounts();
    timers.loadIterationCount2s.stop();
}

void CudaDevice::bootstrapWarpFriendly() {
    timers.bootstrapWarpFriendly.start();
    bootstrap();
    timers.bootstrapWarpFriendly.stop();
}

void CudaDevice::bootstrapNonWarpFriendly() {
    timers.bootstrapNonWarpFriendly.start();
    bootstrap();
    timers.bootstrapNonWarpFriendly.stop();
}

void CudaDevice::copyResultFromGpuToCpu() {
    cudaSetDevice(device_id);

    timers.copyResultFromGpuToCpu.start();
    gpuErrchk( cudaMemcpy(bootstrap_output_host.data(), bootstrap_output_dev, sizeof (float) * n_iterations, cudaMemcpyDeviceToHost) );

    gpuErrchk( cudaFree(bootstrap_output_dev) );
    bootstrap_output_dev = nullptr;

#ifdef DEVICE_SYNCHRONIZE
    gpuErrchk( cudaDeviceSynchronize() );
#endif
    timers.copyResultFromGpuToCpu.stop();
}

std::vector<int> CudaDevice::getIterationCounts(int seeds_from, int seeds_to) {
    cudaSetDevice(device_id);
    iterations_time.start();

    printf("Grid: [%d, %d, %d]\n", grid_dim.x, grid_dim.y, grid_dim.z);
    printf("Block: [%d, %d, %d]\n", block_dim.x, block_dim.y, block_dim.z);

    unsigned int n_seeds_total = seeds_to - seeds_from + 1u;
    unsigned int n_seeds_per_round = 40000000;
    iterations_output_host = std::vector<int>(n_seeds_total, -1);

    std::cout << "input size: " << bootstrap_input_host.size() << std::endl;

    int seed = seeds_from;
    for (unsigned int i = (unsigned int) seeds_from; i < (unsigned int) seeds_to; i += n_seeds_per_round) {
        // A bit hacky, but it works
        n_iterations = std::min(seeds_to - i + 1, n_seeds_per_round);
        gpuErrchk( cudaMalloc((void **) &seeds_dev, sizeof (unsigned int) * n_iterations));

        setCudaTopology();

        seeds_host = std::vector<int>((unsigned int) n_iterations);
        for (unsigned int j = 0; j < (unsigned int) n_iterations; j++) {
            seeds_host[j] = seed;
            seed++;
        }

        copySeedsCpuToGpu();

        gpuErrchk( cudaMalloc((void **) &prng_states_dev, sizeof (curandState_t) * n_iterations) );
        initPrng();

        gpuErrchk( cudaMalloc((void **) &iterations_output_dev, sizeof (int) * n_iterations) );

        if (options.custom_variance) {

            _bootstrap_force_variance_iteration_counts
            <<<grid_dim, block_dim>>>
            (
                bootstrap_input_host.size(),
                iterations_output_dev,
                n_iterations,
                prng_states_dev,
                p,
                options.mu,
                options.sigma
            );
        } else {
            _bootstrap_block_iteration_counts
            <<<grid_dim, block_dim>>>
            (
                bootstrap_input_host.size(),
                iterations_output_dev,
                n_iterations,
                prng_states_dev,
                p
            );
        }
        gpuErrchk( cudaFree(prng_states_dev) );

        gpuErrchk( cudaMemcpy(iterations_output_host.data() + (i - seeds_from), iterations_output_dev, sizeof (int) * n_iterations, cudaMemcpyDeviceToHost) );
        gpuErrchk( cudaFree(iterations_output_dev) );

        std::cout << "Finished iteration #" << ((i - seeds_from)/n_seeds_per_round + 1) << " of " << (n_seeds_total / n_seeds_per_round + 1) << std::endl;
    }

    if (seed != seeds_to + 1) {
        throw Exception("Wrong amount of seeds generated!");
    }

    iterations_output_dev = nullptr;
    prng_states_dev = nullptr;

    iterations_time.stop();

    return iterations_output_host;
}

void CudaDevice::saveEstimators() {
    std::cout << "Saving estimators" << std::endl;
    for (auto node : operation_graph.nodes) {
        auto op = node.second->data;
        op.saveEstimator();
    }
}

void CudaDevice::initPrng() {
    cudaSetDevice(device_id);

    timers.initPrng.start();
    _init_prng
    <<<grid_dim, block_dim>>>
    (
        seeds_dev,
        prng_states_dev,
        n_iterations
    );

    gpuErrchk( cudaFree(seeds_dev) );
    seeds_dev = nullptr;

    timers.initPrng.stop();
}

int CudaDevice::getMaxNIterations() {
    // Calculate maximum number of iterations the devices memory can handle
    // Leave 10MB for storing other stuff.
    // Each iteration uses roughly sizeof (float) bytes, for storing the result of the bootstrapping
    // plus sizeof (curandState_t)
    // plus sizeof (int) (for when we sort on the GPU)
    // plus some buffer (for thrust sorting etc) (experimentally determined)
    gpuErrchk( cudaDeviceSynchronize() );
    cudaSetDevice(device_id);
    gpuErrchk( cudaDeviceSynchronize() );
    size_t free = 0, total = 0;
    cudaMemGetInfo(&free, &total);
    gpuErrchk( cudaDeviceSynchronize() );
    int memPerIteration = (int) (sizeof (curandState_t) + sizeof (float) + sizeof (int) + 8);
    int reserve = 10*1024*1024;
    int deviceMaxNIterations = (int) ((free - reserve) / memPerIteration);
    std::cout << " ===== " << getDeviceName() << " status: ===== " << std::endl;
    std::cout << "Compute capability: " << device_properties.major << "." << device_properties.minor << std::endl;
    std::cout << "Free memory: " << free / (1024*1024) << "MB" << std::endl;
    std::cout << "Bytes per iteration: " << memPerIteration << std::endl;
    std::cout << "Max iterations: " << deviceMaxNIterations << std::endl;
    return deviceMaxNIterations;
}

double CudaDevice::executionPathCost(std::list<Operation> &executionPath, int n_iterations) {
    double cost = 0.0;

    for (Operation &op : executionPath) {
//        std::cout << "Cost of " << op.operationName << ": " << op.cost((int)bootstrap_input_host.size() - 1, n_iterations, p) << std::endl;
        cost += op.cost((int)bootstrap_input_host.size() - 1, n_iterations, p);
    }

    return cost;
}

double CudaDevice::executionPathDCost(std::list<Operation> &executionPath, int n_iterations) {
    double dCost = 0.0;

    for (Operation &op : executionPath) {
//        std::cout << "dCost of " << op.operationName << ": " << op.dCost((int)bootstrap_input_host.size() - 1, n_iterations, p) << std::endl;
        dCost += op.dCost((int)bootstrap_input_host.size() - 1, n_iterations, p);
    }

    return dCost;
}

double CudaDevice::getCost(int n_iterations) {
    auto optimalExecutionPath = findOptimalExecutionPath(n_iterations);
    return executionPathCost(optimalExecutionPath, n_iterations);
}

double CudaDevice::getDCost(int n_iterations) {
    auto optimalExecutionPath = findOptimalExecutionPath(n_iterations);
    return executionPathDCost(optimalExecutionPath, n_iterations);
}
