#include <cudadevicemanager.h>
#include <cudadevice.h>
#include <utils.h>
#include <options.h>


CudaDeviceManager::CudaDeviceManager(std::string iter_counts_file)
    : iter_counts_file(iter_counts_file)
{
    Timer t;
    if (options.load_iter_counts) {
        std::cout << "Loading iteration counts file..." << std::endl;
        t.start();
        read_vector(iter_counts_file, iter_counts);
        std::cout << "Finished loading iteration counts file (" << t.ms() << "ms)." << std::endl;
    } else {
        iter_counts = std::vector<int>();
    }
    json["load_iteration_counts"] = t.ms();
    devices.push_back(std::shared_ptr<Device>(new CudaDevice(options.cuda_device_id, &iter_counts)));
//    for (int i = 0; i < getNumAvailableDevices(); i++) {
//        devices.push_back(std::shared_ptr<Device>(new CudaDevice(i)));
//    }
//    exit(0);
    int runtimeVersion, driverVersion;
    cudaDriverGetVersion(&driverVersion);
    cudaRuntimeGetVersion(&runtimeVersion);
    std::cout << "Using CUDA runtime version: " << runtimeVersion << std::endl;
    std::cout << "Using CUDA driver version: " << driverVersion << std::endl;
    gpuErrchk( cudaDeviceSynchronize() );
    cudaSetDevice(options.cuda_device_id);
    gpuErrchk( cudaDeviceSynchronize() );

    std::cout << "Max iterations: " << std::static_pointer_cast<CudaDevice>(devices[0])->getMaxNIterations() << std::endl;
    gpuErrchk( cudaDeviceSynchronize() );
}

void
CudaDeviceManager::runAsync(std::vector<float> cumulative_input, int n_iterations, float p, Flags flags, int max_threads) {
    auto n_iterations_vec = getLoadBalance(n_iterations);

    future = std::move(
            std::async(std::launch::async, &CudaDeviceManager::run, this, cumulative_input, n_iterations_vec, p, flags)
    );
}

std::vector<std::unique_ptr<Result>>
CudaDeviceManager::run(std::vector<float> cumulative_input, std::vector<int> n_iterations, float p, Flags flags) {

    Timer total_time;
    total_time.start();

    auto cuda_devices = std::vector<std::shared_ptr<CudaDevice>>();
    for (auto &device : devices) {
        cuda_devices.push_back(std::static_pointer_cast<CudaDevice>(device));
    }

    for (int device_id = 0; device_id < (int) cuda_devices.size(); device_id++) {
        cuda_devices[device_id]->setData(cumulative_input, n_iterations[device_id], p, flags);
    }

    std::vector<std::list<Operation>> executionPaths;
    for (int device_id = 0; device_id < (int) cuda_devices.size(); device_id++) {
        auto device = cuda_devices[device_id];
        executionPaths.emplace_back(device->findOptimalExecutionPath(n_iterations[device_id], true));
    }

    nlohmann::json gpu, best_execution_path;
    for (int device_id = 0; device_id < (int) cuda_devices.size(); device_id++) {
        auto executionPath = executionPaths[device_id];

//        std::cout << "Total cost of execution path: " << cuda_devices[device_id]->executionPathCost(executionPath, n_iterations[device_id]) << std::endl;
//        std::cout << "dCost at n_iterations = " << n_iterations[device_id] << ": " << cuda_devices[device_id]->executionPathDCost(executionPath, n_iterations[device_id]) << std::endl;

        for (auto& operation : executionPath) {
            best_execution_path.push_back(operation.operationName);
            operation((int) cumulative_input.size(), n_iterations[device_id], p);
        }
    }

    std::vector<std::unique_ptr<Result>> results;

    for (auto &device_ptr : cuda_devices) {
        results.emplace_back(device_ptr->bootstrapGetResult());
        device_ptr->saveEstimators();
    }

    total_time.stop();

    std::cout << "CUDA devices total time: " << total_time.ms() << "ms" << std::endl;
    Timer total_device_time;
    std::string device_name;
    Timer find_optimal_execution_path_time;
    struct CudaTimers cudaTimers;
    for (auto &res : results) {
        auto cuda_res = static_cast<CudaResult *>(res.get());
        auto timers = cuda_res->bootstrapCudaTimers;
        cudaTimers = timers;
        total_device_time = timers.initialize
                            + timers.generateSeedsCpu
                            + timers.loadIterationCount1s
                            + timers.sortSeedsCpu
                            + timers.copySeedsCpuToGpu
                            + timers.copySeedsAndIterationCountsCpuToGpu

                            + timers.bootstrapWarpFriendly
                            + timers.bootstrapNonWarpFriendly

                            + timers.sortSeedsGpu
                            + timers.generateSeedsGpu
                            + timers.copySeedsGpuToCpu
                            + timers.loadIterationCount2s
                            + timers.copyIterationCountsCpuToGpu

                            + timers.copyResultFromGpuToCpu
                            + timers.finalize;
        std::cout << res->device_name << " total time: " << total_device_time.ms() << "ms" << std::endl;
        device_name = res->device_name;
        find_optimal_execution_path_time = timers.getOptimalExecutionPath;
    }

    nlohmann::json gpu_timers;
    gpu_timers["Initialize"] = cudaTimers.initialize.ms();
    gpu_timers["Generate seeds CPU"] = cudaTimers.generateSeedsCpu.ms();
    gpu_timers["Sort seeds on the CPU"] = cudaTimers.sortSeedsCpu.ms();
    gpu_timers["Get iteration counts 1"] = cudaTimers.loadIterationCount1s.ms();
    gpu_timers["Get iteration counts 2"] = cudaTimers.loadIterationCount2s.ms();
    gpu_timers["Copy seeds from GPU to CPU"] = cudaTimers.copySeedsGpuToCpu.ms();
    gpu_timers["Copy seeds from CPU to GPU"] = cudaTimers.copySeedsCpuToGpu.ms();
    gpu_timers["Copy iteration counts from CPU to GPU"] = cudaTimers.copyIterationCountsCpuToGpu.ms();
    gpu_timers["Copy seeds and iteration counts from CPU to GPU"] = cudaTimers.copySeedsAndIterationCountsCpuToGpu.ms();
    gpu_timers["Generate seeds GPU"] = cudaTimers.generateSeedsGpu.ms();
    gpu_timers["Sort seeds on the GPU"] = cudaTimers.sortSeedsGpu.ms();
    gpu_timers["Initialize PRNG 1"] = cudaTimers.initPrng.ms();
    gpu_timers["Initialize PRNG 2"] = cudaTimers.initPrng.ms();
    gpu_timers["Warp-friendly bootstrap"] = cudaTimers.bootstrapWarpFriendly.ms();
    gpu_timers["Non warp-friendly bootstrap"] = cudaTimers.bootstrapNonWarpFriendly.ms();
    gpu_timers["Copy result from GPU to CPU"] = cudaTimers.copyResultFromGpuToCpu.ms();
    gpu_timers["Finalize"] = cudaTimers.finalize.ms();

    gpu["execution_time"] = total_device_time.ms();
    gpu["best_execution_path"] = best_execution_path;
    gpu["best_execution_path_ms"] = find_optimal_execution_path_time.ms();
    gpu["timers"] = gpu_timers;
    gpu["device_name"] = device_name;
    json["gpu"] = gpu;

    return results;
}

int
CudaDeviceManager::getNumAvailableDevices() {
    int n_devices;

    cudaGetDeviceCount(&n_devices);

    return 1; //n_devices;
}

std::vector<int>
CudaDeviceManager::getLoadBalance(int n_iterations) {
    int n_devices = getNumAvailableDevices();

    double frac = 1.0 / n_devices;

    std::vector<int> n_iterations_vec((unsigned long) n_devices);
    int total = 0;
    for (int i = 0; i < n_devices - 1; i++) {
        int cur_iter = (int) (frac * n_iterations);
        total += cur_iter;
        n_iterations_vec[i] = cur_iter;
    }
    n_iterations_vec[n_devices-1] = n_iterations - total;
    return n_iterations_vec;
}
