#include <estimator.h>
#include <utils.h>


Estimator::Estimator(std::string savePath)
    : need_rebuild(false),
      basePath(savePath),
      alpha(0.1)
{
//    std::cout << "Created Estimator with save path " << savePath << std::endl;
    if (file_exists(getDataTablePath())) {
//        std::cout << "Loading datatable from " << getDataTablePath() << std::endl;
        dataTable_ptr = std::unique_ptr<SPLINTER::DataTable>(new SPLINTER::DataTable(getDataTablePath()));
//        std::cout << "Num samples: " << dataTable_ptr->getNumSamples() << std::endl;
        need_rebuild = true;
    }
    else {
        dataTable_ptr = std::unique_ptr<SPLINTER::DataTable>(new SPLINTER::DataTable(false));
    }
//    std::cout << "Datatable input dimension: " << dataTable_ptr->getNumVariables() << std::endl;
    bSpline_ptr = std::unique_ptr<SPLINTER::BSpline>(nullptr);
}

Estimator::~Estimator() {

}

void Estimator::train(int input_size, int n_iterations, float p, double cost) {
    std::vector<double> x(3);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    x[2] = (double) p;
    dataTable_ptr->addSample(x, cost);
    need_rebuild = true;
}

void Estimator::train(int input_size, int n_iterations, double cost) {
    std::vector<double> x(2);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    dataTable_ptr->addSample(x, cost);
    need_rebuild = true;
}

void Estimator::train(int n_iterations, double cost) {
//    std::cout << "Samples before training " << getBSplinePath() << ":" << std::endl;
//    for (auto it = dataTable_ptr->cbegin(); it != dataTable_ptr->cend(); it++) {
//        std::cout << it->getX()[0] << " = " << it->getY() << std::endl;
//    }
    dataTable_ptr->addSample((double) n_iterations, cost);
    need_rebuild = true;
//    std::cout << "Samples after training " << getBSplinePath() << ":" << std::endl;
//    for (auto it = dataTable_ptr->cbegin(); it != dataTable_ptr->cend(); it++) {
//        std::cout << it->getX()[0] << " = " << it->getY() << std::endl;
//    }
}

void Estimator::save() {
//    std::cout << "Saving datatable to " << getDataTablePath() << std::endl;

    dataTable_ptr->save(getDataTablePath());

    rebuildIfNecessary();
    if (bSpline_ptr != nullptr) {
        // Save the BSpline too, so it can be loaded in Python for further analysis if needed
        bSpline_ptr->save(getBSplinePath());
    }
}

double Estimator::estimate(int n_iterations) {
    rebuildIfNecessary();

    std::vector<double> x(1);
    x[0] = (double) n_iterations;
    return bSpline_ptr->eval(x);
}

double Estimator::estimate(int input_size, int n_iterations) {
    rebuildIfNecessary();

    std::vector<double> x(2);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    return bSpline_ptr->eval(x);
}

double Estimator::estimate(int input_size, int n_iterations, float p) {
    rebuildIfNecessary();

    std::vector<double> x(3);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    x[2] = (double) p;
    return bSpline_ptr->eval(x);
}

double Estimator::dEstimate(int n_iterations) {
    rebuildIfNecessary();

    std::vector<double> x(1);
    x[0] = (double) n_iterations;
    auto jacobian = bSpline_ptr->evalJacobian(x);
    return jacobian[0];
}

double Estimator::dEstimate(int input_size, int n_iterations) {
    rebuildIfNecessary();

    std::vector<double> x(2);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    auto jacobian = bSpline_ptr->evalJacobian(x);
    return jacobian[1]; // Return the second element, we want the derivative of the n_iterations dimension
}

double Estimator::dEstimate(int input_size, int n_iterations, float p) {
    rebuildIfNecessary();

    std::vector<double> x(3);
    x[0] = (double) input_size;
    x[1] = (double) n_iterations;
    x[2] = (double) p;
    auto jacobian = bSpline_ptr->evalJacobian(x);
    return jacobian[1]; // Return the second element, we want the derivative of the n_iterations dimension
}

void Estimator::setAlpha(double newAlpha) {
    need_rebuild = true;
    alpha = newAlpha;
}

void Estimator::rebuildIfNecessary() {
    if (canEstimate() && (bSpline_ptr == nullptr || need_rebuild)) {
//        std::cout << "Building B-spline from samples for estimator " << getBSplinePath() << ":" << std::endl;
//        for (auto it = dataTable_ptr->cbegin(); it != dataTable_ptr->cend(); it++) {
//            std::cout << it->getX()[0] << " = " << it->getY() << std::endl;
//        }
        bSpline_ptr = std::unique_ptr<SPLINTER::BSpline>(
                SPLINTER::BSpline::Builder(*dataTable_ptr)
                        .degree(3)
                        .alpha(alpha)
                        .smoothing(SPLINTER::BSpline::Smoothing::PSPLINE)
                        .numBasisFunctions(30)
                        .knotSpacing(SPLINTER::BSpline::KnotSpacing::EQUIDISTANT)
                        .build()
                        .clone()
        );
        need_rebuild = false;
    }
}

std::string Estimator::getDataTablePath() {
    return basePath + ".datatable";
}

std::string Estimator::getBSplinePath() {
    return basePath + ".bspline";
}

bool Estimator::canEstimate() {
    int dim = (int) dataTable_ptr->getNumVariables();
    int minNumSamples = 1;
    // Should just use 4**dim
    for (int i = 0; i < dim; i++) {
        minNumSamples *= 4;
    }
//    auto grid = dataTable_ptr->getGrid();
//    for (auto &set : grid) {
//        for (auto i : set) {
//            std::cout << i << " ";
//        }
//        std::cout << std::endl;
//    }
    return dataTable_ptr->isGridComplete() && (int) dataTable_ptr->getNumSamples() >= minNumSamples;
}
