#include <device_launch_parameters.h>
#include <math_functions.h>
#include <curand_kernel.h>
#include <cstdio>
#include <limits>


__device__ static int _normal_random(curandState_t *state, float mu, float sigma) {
    float u1 = curand_uniform(state);
    float u2 = curand_uniform(state);
    float two_pi = (float) (2.0*3.14159265358979323846);
    float z0 = (float) (sqrt(-2.0 * logf(u1)) * cosf(two_pi * u2));
    return (int) (z0*sigma + mu);
}

// http://math.stackexchange.com/questions/485448/prove-the-way-to-generate-geometrically-distributed-random-numbers
__device__ static unsigned int _geometric(curandState_t *state, float p) {
    float random_num = curand_uniform(state);
    return (unsigned int) (logf(random_num) / logf(1 - p)) + 1;
}

// Draw uniform in [from, to)
__device__ static int _uniform(curandState_t *randState, int from, int to) {
    int delta = to - from;
    return from + curand(randState) % delta;
}

__global__ void _init_prng(int *seeds, curandState_t *states, int n_bootstrap_iterations) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    if (id < n_bootstrap_iterations) {
        curand_init((unsigned int) seeds[id], 0, 0, &(states[id]));
    }
}

__global__ void _bootstrap_block(float *cumulativeInput, int cumulativeInputSize, float *out, int n_bootstrap_iterations, curandState_t *states, float p) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    extern __shared__ float c[];

    int inputSize = cumulativeInputSize - 1;
    int n_threads_block_dim_x = min(blockDim.x, n_bootstrap_iterations);
    int n = cumulativeInputSize / n_threads_block_dim_x;
    if (cumulativeInputSize % n_threads_block_dim_x != 0) {
        n++;
    }
    for (int i = 0; i < n; ++i) {
        int idx = threadIdx.x + n_threads_block_dim_x*i;
        if (idx < cumulativeInputSize) {
            c[idx] = cumulativeInput[idx];
        }
    }

    __syncthreads();

    curandState_t *randState = &states[id];

    if (id < n_bootstrap_iterations) {
        float sum = 0.0f;

        for (int available = inputSize; available > 0; ) {
            int start = _uniform(randState, 0, inputSize);

            int block_length = min(_geometric(randState, p), available);

            available -= block_length;

            int end = start + block_length;

            int end_clamped = min(end, inputSize);

            int end_wrapped_around = max(0, end - inputSize);

            sum += c[end_clamped] - c[start] + c[end_wrapped_around];
        }

        out[id] = sum / inputSize;
    }
}

__global__ void _bootstrap_block_long(float *cumulativeInput, int cumulativeInputSize, float *out, int n_bootstrap_iterations, curandState_t *states, float p) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    int inputSize = cumulativeInputSize - 1;

    curandState_t *randState = &states[id];

    if (id < n_bootstrap_iterations) {
        float sum = 0.0f;

        for (int available = inputSize; available > 0; ) {
            int start = _uniform(randState, 0, inputSize);

            int block_length = min(_geometric(randState, p), available);

            available -= block_length;

            int end = start + block_length;

            int end_clamped = min(end, inputSize);

            int end_wrapped_around = max(0, end - inputSize);

            sum += cumulativeInput[end_clamped] - cumulativeInput[start] + cumulativeInput[end_wrapped_around];
        }

        out[id] = sum / inputSize;
    }
}

__global__ void _bootstrap_block_iteration_counts(int cumulativeInputSize, int *out, int n_bootstrap_iterations, curandState_t *states, float p) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    int inputSize = cumulativeInputSize - 1;

    curandState_t *randState = &states[id];

    if (id < n_bootstrap_iterations) {
        int iter = 0;
        for (int available = inputSize; available > 0; iter++) {
            int start = _uniform(randState, 0, inputSize);

            int block_length = min(_geometric(randState, p), available);

            available -= block_length;

            // Apparently, even though they're not used, these cannot be removed...
            // Maybe an incorrect optimization by the compiler?
            int end = start + block_length;

            int end_clamped = min(end, inputSize);

            int end_wrapped_around = max(0, end - inputSize);
        }

        out[id] = iter;
    }
}

__global__ void _bootstrap_force_variance(float *cumulativeInput, int cumulativeInputSize,
                                          float *out, int n_bootstrap_iterations,
                                          curandState_t *states, float p, float mu, float sigma) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    extern __shared__ float c[];

    int inputSize = cumulativeInputSize - 1;
    int n_threads_block_dim_x = min(blockDim.x, n_bootstrap_iterations);
    int n = cumulativeInputSize / n_threads_block_dim_x;
    if (cumulativeInputSize % n_threads_block_dim_x != 0) {
        n++;
    }
    for (int i = 0; i < n; ++i) {
        int idx = threadIdx.x + n_threads_block_dim_x*i;
        if (idx < cumulativeInputSize) {
            c[idx] = cumulativeInput[idx];
        }
    }

    __syncthreads();

    curandState_t *randState = &states[id];

    if (id < n_bootstrap_iterations) {
        int n_iterations = max(0, _normal_random(randState, mu, sigma));

        float sum = 0.0f;

        int available = inputSize;
        for (int i = 0; i < n_iterations; i++) {
            int start = _uniform(randState, 0, inputSize);

            int block_length = min(_geometric(randState, p), inputSize);

            available -= block_length;

            int end = start + block_length;

            int end_clamped = min(end, inputSize);

            int end_wrapped_around = max(0, end - inputSize);

            sum += c[end_clamped] - c[start] + c[end_wrapped_around];
        }

        out[id] = sum / inputSize;
    }
};

__global__ void _bootstrap_force_variance_iteration_counts(
        int cumulativeInputSize,
        int *out,
        int n_bootstrap_iterations,
        curandState_t *states,
        float p,
        float mu,
        float sigma
) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    curandState_t *randState = &states[id];

    if (id < n_bootstrap_iterations) {
        out[id] = max(0, _normal_random(randState, mu, sigma));
    }
}

__global__ void _generate_seeds(int global_seed, int *out, int n_iterations) {
    int block_id = blockIdx.y * gridDim.x + blockIdx.x;
    int id = block_id * blockDim.x * blockDim.y + threadIdx.y * blockDim.x + threadIdx.x;

    if (id < n_iterations) {
        curandState_t rand_state;
        // Intentional unsigned integer under/overflow
        curand_init((unsigned int) (global_seed + id), 0, 0, &rand_state);
        out[id] = (int) (curand(&rand_state) / 2);
    }
}
