#include <options.h>
#include <stdlib.h>
#include <stdio.h>
#include <cstring>
#include <iostream>
#include <fstream>
#include "bootstrap.h"
#include <serializer.h>
#include <histogram.h>
#include <json/json.h>
#include <iomanip>
#include <cudadevice.h>
#include <experimental/filesystem>


enum RETURN_CODES {
    SUCCESS = 0,
    INVALID_GPU_FRACTION = 1,
    WRONG_USAGE = 2,
    INVALID_P = 3,
};

float randfloat(float from, float to) {
    return from + ((rand() / (float)RAND_MAX) * (to - from));
}


std::vector<float> read_input(int n, float rand_from, float rand_to) {
    auto input = std::vector<float>();
    for (int i = 0; i < n; ++i) {
        input.push_back(randfloat(rand_from, rand_to));
    }
    return input;
}


int *histogram(float *numbers, int n_numbers, int n_buckets) {
    // Bucket #1: 0-1. Bucket #10: 9-10
    int *buckets = (int *) malloc(sizeof (int) * n_buckets);
    for (int i = 0; i < n_buckets; ++i) {
        buckets[i] = 0;
    }
    for (int i = 0; i < n_numbers; ++i) {
        int bucket_number = (int) (numbers[i] + 0.5f);
        if (bucket_number >= n_buckets) {
            printf("bucket_number >= n_buckets! bucket_number: %d\n", bucket_number);
        }
        else {
            buckets[bucket_number]++;
        }
    }
    return buckets;
}

std::string series_length_option("-l");
std::string n_iterations_option("-n");
std::string p_option("-p");
std::string gpu_fraction_option("--gpu-fraction");
std::string sort_option("--sort");
std::string sort_warp_friendly_option("--warp-friendly");
std::string generate_iter_counts_option("--generate-iter-counts");
std::string output_file_option("--output-file");
std::string histogram_from_option("--histogram-from");
std::string histogram_to_option("--histogram-to");
std::string histogram_n_buckets("--histogram-n-buckets");
std::string generate_iter_histogram_option("--generate-iter-histogram");
std::string generate_seeds_option("--generate-seeds");
std::string force_path_option("--force-path");
std::string storage_path_option("--storage-path");
std::string calc_iter_variance_option("--calc-iter-variance");
std::string force_variance_option("--force-variance");
std::string device_id_option("--device-id");
std::string calc_cv_option("--calc-cv");

void print_usage(int argc, char **argv) {
    std::cout << "Usage: " << argv[0] << " "
              << series_length_option << "=SERIES_LENGTH "
              << n_iterations_option << "=N_ITERATIONS "
              << p_option << "=P_VALUE"
              << std::endl;
    std::cout << "Additional options:" << std::endl;
    std::cout << sort_option << "=CPU|GPU                             # Sort the seeds (on the CPU or GPU) based on the number of iterations on the GPU" << std::endl;
    std::cout << gpu_fraction_option << "=0.5                         # Force the fraction of iterations on the GPU ([0.0, 1.0])" << std::endl;
    std::cout << sort_warp_friendly_option << "                            # Ensure that all threads in each warp have the same iteration count" << std::endl;
    std::cout << generate_iter_counts_option << "=output_file_option  # Calculate and output the seed => iteration_count for the time series of length specified earlier" << std::endl;
    std::cout << output_file_option << "=output_file.json             # File to output result and timings of the run to." << std::endl;
    std::cout << histogram_from_option << "=0.0                       # Left interval border of first bucket in histogram" << std::endl;
    std::cout << histogram_to_option << "=1.0                         # Right interval border of last bucket in histogram" << std::endl;
    std::cout << histogram_n_buckets << "=10                   # Number of buckets in histogram" << std::endl;
    std::cout << generate_iter_histogram_option << "=output_file.json # Generate iteration count histogram and output JSON to file" << std::endl;
    std::cout << generate_seeds_option << "=CPU|GPU                   # Where to generate the seeds" << std::endl;
    std::cout << force_path_option << "=warp_friendly                 # Force execution through operation" << std::endl;
    std::cout << storage_path_option << "=/path/to/directory          # Set directory to store estimators and seeds_to_iterations" << std::endl;
    exit(WRONG_USAGE);
}


float get_float_option(const InputParser &inputParser, std::string option) {
    auto option_str = inputParser.getCmdOption(option);
    char *end = const_cast<char *>(option_str.c_str()) + option_str.size();
    return (float) strtod(option_str.c_str(), &end);
}

int get_int_option(const InputParser &inputParser, std::string option) {
    auto option_str = inputParser.getCmdOption(option);
    return atoi(option_str.c_str());
}


int main(int argc, char **argv) {
    srand((unsigned int) time(NULL));

    using path = std::experimental::filesystem::path;

    try {
        InputParser inputParser(argc, argv);

        Flags flags = Flags::NONE;

        if (!inputParser.cmdOptionExists(n_iterations_option)
            || !inputParser.cmdOptionExists(series_length_option)
            || !inputParser.cmdOptionExists(p_option)) {
            print_usage(argc, argv);
        }

        int series_length = get_int_option(inputParser, series_length_option);
        int n_iterations = get_int_option(inputParser, n_iterations_option);
        auto output_file = std::string("");

        float p = get_float_option(inputParser, p_option);
        if (!(0.0f < p && p <= 1.0f)) {
            std::cout << "p must be in the interval (0.0, 1.0]!" << std::endl;
            exit(INVALID_P);
        }

        options.forceExecutionPath = inputParser.getCmdOption(force_path_option);
        options.storageDir = std::experimental::filesystem::current_path();
        if (inputParser.cmdOptionExists(storage_path_option)) {
            options.storageDir = inputParser.getCmdOption(storage_path_option);
        }

        options.iterCountsDir = options.storageDir / path("seeds_to_iterations");
        options.iterCountsFilePath = options.iterCountsDir / path("l" + inputParser.getCmdOption(series_length_option) + "_p" + inputParser.getCmdOption(p_option));
        options.estimatorsDir = options.storageDir / path("estimators") / path("l" + inputParser.getCmdOption(series_length_option) + "_p" + inputParser.getCmdOption(p_option));

        std::experimental::filesystem::create_directories(options.storageDir);
        std::experimental::filesystem::create_directories(options.iterCountsDir);
        std::experimental::filesystem::create_directories(options.estimatorsDir);

        std::cout << "Storage dir: " << options.storageDir << std::endl;
        std::cout << "Seeds to iterations dir: " << options.iterCountsDir << std::endl;
        std::cout << "Seeds to iterations file: " << options.iterCountsFilePath << std::endl;
        std::cout << "Estimators dir: " << options.estimatorsDir << std::endl;

        options.load_iter_counts = inputParser.getCmdOption(force_path_option) != std::string("non_warp_friendly");

        options.custom_variance = inputParser.cmdOptionExists(force_variance_option);
        options.cuda_device_id = 0;
        if (inputParser.cmdOptionExists(device_id_option)) {
            options.cuda_device_id = get_int_option(inputParser, device_id_option);
        }

        if (options.custom_variance) {
            if (!inputParser.cmdOptionExists("--mu") || !inputParser.cmdOptionExists("--sigma")) {
                std::cout << "Both --mu and --sigma must be specified!" << std::endl;
                exit(1);
            }
            options.mu = (float) get_int_option(inputParser, "--mu");
            options.sigma = (float) get_int_option(inputParser, "--sigma");
            std::cout << "Forcing workload with mu=" << options.mu << " and sigma=" << options.sigma << std::endl;
        }

        if (inputParser.cmdOptionExists(output_file_option)) {
            output_file = inputParser.getCmdOption(output_file_option);
        }

        if (inputParser.cmdOptionExists(calc_cv_option)) {
            auto output_file_name = inputParser.getCmdOption(calc_cv_option);
            std::cout << "Calculating coefficient of variation for series length " << series_length;
            std::cout << " and outputting result to \"" << output_file_name << "\"" << std::endl;

            // From and to has no impact on iteration counts
            std::vector<float> fakeInput((unsigned int)series_length+1);
            CudaDevice cudaDevice0(options.cuda_device_id, nullptr);
            std::cout << "Using " << cudaDevice0.getDeviceName() << " for generating seeds to iteration counts" << std::endl;
            cudaDevice0.setData(fakeInput, n_iterations, p, flags);

            int seeds_from = 0;
            int seeds_to = std::numeric_limits<int>::max() / 2;
            unsigned int num_seeds = seeds_to - seeds_from + 1u;
            std::cout << "seeds_from: " << seeds_from << std::endl;
            std::cout << "seeds_to: " << seeds_to << std::endl;
            std::cout << "Generating " << num_seeds << " seed => iteration_counts mappings." << std::endl;

            auto iteration_counts0 = std::move(cudaDevice0.getIterationCounts(seeds_from, seeds_to));

            unsigned long tot = 0;
            for (int iter_count : iteration_counts0) {
                tot += iter_count;
            }
            std::cout << "Total: " << tot << std::endl;
            double average = (double) tot / (double) iteration_counts0.size();
            double var_sum = 0.0;
            for (int iter_count : iteration_counts0) {
                var_sum += ((double) iter_count - average) * ((double) iter_count - average);
            }
            double variance = var_sum / iteration_counts0.size();
            double cv = sqrt(variance) / average;

            std::cout << "Outputting to " << output_file_name << std::endl;
            if (output_file_name.compare("") != 0) {
                if (file_exists(output_file_name)) {
                    throw Exception("Output file already exists!");
                }
                nlohmann::json j;
                std::ofstream out(output_file_name);
                j["l"] = series_length;
                j["p"] = p;
                j["variance"] = variance;
                j["average"] = average;
                j["cv"] = cv;
                out << std::setw(4) << j;
                out.close();
            }
        }

        else if (inputParser.cmdOptionExists(generate_iter_histogram_option)) {
            auto output_file_name = inputParser.getCmdOption(generate_iter_histogram_option);
            std::cout << "Generating histogram of iteration counts for series length " << series_length;
            std::cout << " and outputting result to \"" << output_file_name << "\"" << std::endl;

            // From and to has no impact on iteration counts
            std::vector<float> fakeCumulativeInput((unsigned int)series_length+1);
            CudaDevice cudaDevice0(options.cuda_device_id, nullptr);
            std::cout << "Using " << cudaDevice0.getDeviceName() << " for generating seeds to iteration counts" << std::endl;
            cudaDevice0.setData(fakeCumulativeInput, n_iterations, p, flags);

            int seeds_from = 0;
            int seeds_to = std::numeric_limits<int>::max() / 4;
            unsigned int num_seeds = seeds_to - seeds_from + 1u;
            std::cout << "seeds_from: " << seeds_from << std::endl;
            std::cout << "seeds_to: " << seeds_to << std::endl;
            std::cout << "Generating " << num_seeds << " seed => iteration_counts mapping." << std::endl;

            auto iteration_counts0 = std::move(cudaDevice0.getIterationCounts(seeds_from, seeds_to));

            double tot = 0.0;
            for (auto iter_count : iteration_counts0) {
                tot += iter_count;
            }
            std::cout << "Average: " << tot / iteration_counts0.size() << std::endl;

            auto histogram = Histogram<int, double>(std::move(iteration_counts0));
            histogram.reorganize(histogram.maxVal - histogram.minVal, histogram.minVal, histogram.maxVal);
            histogram.normalize();
//            std::cout << "normalized " << std::endl;
//            for (int i = 0; i < histogram.getNumBuckets(); i++) {
//                std::cout << histogram.getBucketInterval(i).first << " - " << histogram.getBucketInterval(i).second << ": " << histogram[i] << std::endl;
//            }

            std::cout << "Outputting to " << output_file_name << std::endl;
            if (output_file_name.compare("") != 0) {
                if (file_exists(output_file_name)) {
                    throw Exception("Output file already exists!");
                }
                nlohmann::json j;
                std::ofstream out(output_file_name);
                j["l"] = series_length;
                j["p"] = p;
                j["n_seeds"] = num_seeds;
                j["histogram"] << histogram;
                out << std::setw(4) << j;
                out.close();
            }
        }

        else if (inputParser.cmdOptionExists(calc_iter_variance_option)) {
            auto output_file_name = inputParser.getCmdOption(calc_iter_variance_option);
            std::cout << "Generating variance of iteration counts for series length " << series_length;
            std::cout << " and outputting result to \"" << output_file_name << "\"" << std::endl;

            // From and to has no impact on iteration counts
            std::vector<float> fakeInput((unsigned int)series_length+1);
            CudaDevice cudaDevice0(cuda_device, nullptr);
            std::cout << "Using " << cudaDevice0.getDeviceName() << " for generating seeds to iteration counts" << std::endl;
            cudaDevice0.setData(fakeInput, n_iterations, p, flags);

            int seeds_from = 0;
            int seeds_to = std::numeric_limits<int>::max() / 4;
            unsigned int num_seeds = seeds_to - seeds_from + 1u;
            std::cout << "seeds_from: " << seeds_from << std::endl;
            std::cout << "seeds_to: " << seeds_to << std::endl;
            std::cout << "Generating " << num_seeds << " seed => iteration_counts mapping." << std::endl;

            auto iteration_counts0 = std::move(cudaDevice0.getIterationCounts(seeds_from, seeds_to));

            unsigned long tot = 0;
            for (int iter_count : iteration_counts0) {
                tot += iter_count;
            }
            std::cout << "Total: " << tot << std::endl;
            double average = (double) tot / (double) iteration_counts0.size();
            double var_sum = 0.0;
            for (int iter_count : iteration_counts0) {
                var_sum += ((double) iter_count - average) * ((double) iter_count - average);
            }
            double variance = var_sum / iteration_counts0.size();

            std::cout << "Outputting to " << output_file_name << std::endl;
            if (output_file_name.compare("") != 0) {
                if (file_exists(output_file_name)) {
                    throw Exception("Output file already exists!");
                }
                nlohmann::json j;
                std::ofstream out(output_file_name);
                j["l"] = series_length;
                j["p"] = p;
                j["n_seeds"] = num_seeds;
                j["variance"] = variance;
                j["average"] = average;
                out << std::setw(4) << j;
                out.close();
            }
        }

        else if ((!file_exists(options.iterCountsFilePath) || inputParser.cmdOptionExists(generate_iter_counts_option)) && options.load_iter_counts) {
            auto output_file_name = options.iterCountsFilePath.generic_string();
            std::cout << "Generating seeds => iteration_count mapping for series of length " << series_length;
            std::cout << " and outputting result to " << options.iterCountsFilePath << std::endl;

            if (file_exists(options.iterCountsFilePath)) {
                throw Exception("File already exists!");
            }

            // From and to has no impact on iteration counts
            std::vector<float> fakeInput((unsigned int)series_length+1);
            CudaDevice cudaDevice0(0, nullptr);
            cudaDevice0.setData(fakeInput, n_iterations, p, flags);

            int seeds_from = 0;
            int seeds_to = std::numeric_limits<int>::max();
            unsigned int num_seeds = seeds_to - seeds_from + 1u;
            std::cout << "seeds_from: " << seeds_from << std::endl;
            std::cout << "seeds_to: " << seeds_to << std::endl;
            std::cout << "Generating " << num_seeds << " seed => iteration_counts mappings." << std::endl;

            auto iteration_counts0 = std::move(cudaDevice0.getIterationCounts(seeds_from, seeds_to));
            write_vector(output_file_name, iteration_counts0);
        }

        else {
            if (inputParser.cmdOptionExists(sort_option)) {
                std::string sort = inputParser.getCmdOption(sort_option);

                std::transform(sort.begin(), sort.end(), sort.begin(), ::tolower);
                if (sort == "gpu") {
                    flags |= Flags::SORT_GPU;
                }
                else if (sort == "cpu") {
                    flags |= Flags::SORT_CPU;
                }
                else {
                    print_usage(argc, argv);
                }
            }

            if (inputParser.cmdOptionExists(generate_seeds_option)) {
                std::string generate_seeds = inputParser.getCmdOption(generate_seeds_option);
                std::transform(generate_seeds.begin(), generate_seeds.end(), generate_seeds.begin(), ::tolower);
                std::cout << generate_seeds << std::endl;
                if (generate_seeds == "gpu") {
                    flags |= Flags::GENERATE_SEEDS_GPU;
                }
                else if (generate_seeds == "cpu") {
                    flags |= Flags::GENERATE_SEEDS_CPU;
                }
                else {
                    print_usage(argc, argv);
                }
            }

            if (inputParser.cmdOptionExists(sort_warp_friendly_option)) {
                flags |= Flags::GROUP_INTO_WARPS;
            }

            float gpu_fraction = nanf("");
            if (inputParser.cmdOptionExists(gpu_fraction_option)) {
                gpu_fraction = get_float_option(inputParser, gpu_fraction_option);
                if (!(0.0f <= gpu_fraction && gpu_fraction <= 1.0f)) {
                    std::cout << "gpu_fraction must be in the interval [0.0, 1.0]!" << std::endl;
                    exit(INVALID_GPU_FRACTION);
                }
                flags |= Flags::MANUAL_LOAD_BALANCE;
            }

            printf("Input size: %d\n", series_length);
            printf("Bootstrap iterations: %d\n", n_iterations);
            printf("p value geometric distribution: %f\n", p);
            std::cout << "GPU fraction: " << gpu_fraction << std::endl;
            std::cout << "Sort warp friendly: " << ((flags & Flags::GROUP_INTO_WARPS) ? "yes" : "no") << std::endl;
            std::cout << "Flags: " << flags << std::endl;

            float
                    random_from = 0.0f,
                    random_to = 10.0f;

            auto input = read_input(series_length, random_from, random_to);
            std::cout << "min: " << *std::min_element(input.cbegin(), input.cend()) << std::endl;
            std::cout << "max: " << *std::max_element(input.cbegin(), input.cend()) << std::endl;

            double d = 0.0;
            for (auto f : input) {
                d += f;
            }
            std::cout << "avg: " << d / input.size() << std::endl;

            Bootstrap bootstrap(options.iterCountsFilePath);
            auto result = bootstrap.bootstrap(input, n_iterations, p, flags, gpu_fraction);

            double avg = 0.0;
            for (auto f : result) {
                avg += f;
            }
            avg /= result.size();
            double temp = 0.0;
            for (auto f : result) {
                temp += (f - avg) * (f - avg);
            }
            std::cout << "Variance: " << temp / result.size() << std::endl;

            std::cout << "Result size: " << result.size() << std::endl;
            Histogram<float, float> histogram(result);
            histogram.reorganize(5, 4.96935f-0.2f, 4.96935f+0.2f);
            histogram.normalize();

    //        if (output_file.compare("") != 0) {
    //            if (file_exists(output_file)) {
    //                throw Exception("Output file already exists!");
    //            }
    //            nlohmann::json j;
    //            std::ofstream out(output_file);
    //            j["histogram"] << histogram;
    //            j["stats"] << bootstrap;
    //            out << std::setw(4) << j;
    //            out.close();
    //        }

            std::cout << "Buckets start: " << histogram.getBucketsStart() << std::endl;
            std::cout << "Buckets end:  " << histogram.getBucketsEnd() << std::endl;
            for (int i = -1; i < histogram.getNumBuckets() + 1; i++) {
                std::cout << histogram.getBucketInterval(i).first << "-" << histogram.getBucketInterval(i).second << " = " << histogram[i] << std::endl;
            }
            std::cout << std::endl;
            json["histogram"] << histogram;

            if (output_file != "") {
                if (file_exists(output_file)) {
                    std::cout << "Error: Output file already exists: \"" << output_file << "\"" << std::endl;
                    exit(1);
                }
                std::ofstream out(output_file);
                out << std::setw(4) << json;
                out.close();
            }
    //
    //        std::cout << "Absolute total GPU time: " << bootstrap.gpu_time.s() << std::endl;
    //        std::cout << "Average GPU time: " << bootstrap.gpu_time.s() / 10 << std::endl;
    //        std::cout << "Sum of timers: " <<
    //                  (
    //                          bootstrap.seeds_sort_time
    //                          + bootstrap.generate_seeds_time
    //                          + bootstrap.bootstrapCUDA->getBootstrapTimer()
    //                  ).s() << "s" << std::endl;
        }
    }
    catch (const Exception &e) {
        std::cout << "Got Exception: " << e.what() << std::endl;
    }

    return SUCCESS;
}
