#include <operation.h>
#include "options.h"
#include <experimental/filesystem>


unsigned int Operation::next_id = 0;


Operation::Operation(const std::string &operationName)
    :
        id(Operation::next_id++),
        operationName(operationName),
        dummy(true),
        inputSizeDependent(false),
        pDependent(false)
{

}

Operation::Operation(const std::string &deviceName,
                     const std::string id,
                     const std::string &operationName,
                     memberFunc opFunc,
                     CudaDevice *self,
                     bool inputSizeDependent,
                     bool pDependent)
        :
        id(Operation::next_id++),
        operationName(operationName),
        dummy(false),
        operation(opFunc),
        self(self),
        forceLowEstimate(false),
        inputSizeDependent(inputSizeDependent),
        pDependent(pDependent)
{
    forceLowEstimate = options.forceExecutionPath == id;
    if (forceLowEstimate) {
        std::cout << "Forcing execution through " << operationName << std::endl;
    }
    using std::experimental::filesystem::path;
    path estimatorSaveDir = options.estimatorsDir / path(deviceName);
    std::experimental::filesystem::create_directories(estimatorSaveDir);
    estimator_ptr = std::shared_ptr<Estimator>(new Estimator(estimatorSaveDir / path(operationName)));
}

Operation::~Operation() {

}

double Operation::cost(int input_size, int n_iterations, float p) {
    if (dummy) {
        return 0.0;
    }

    if (forceLowEstimate) {
//        std::cout << "Returning very low estimate from " << operationName << std::endl;
        return std::numeric_limits<float>::lowest();
    }

    if (!estimator_ptr->canEstimate()) {
//        std::cout << "Estimator for \"" << operationName << "\" not sufficiently trained yet, returning 0.0 as cost." << std::endl;
        // Something cannot ever take less than no time
        return 0.0;
    }

    double splineEstimate = 0.0;
    if (pDependent) {
        splineEstimate = estimator_ptr->estimate(input_size, n_iterations, p);
    }
    else if (inputSizeDependent) {
        splineEstimate = estimator_ptr->estimate(input_size, n_iterations);
    }
    else {
        splineEstimate = estimator_ptr->estimate(n_iterations);
    }
    if (splineEstimate < 0.0) {
        splineEstimate = 0.0;
    }
    return splineEstimate;
}

double Operation::dCost(int input_size, int n_iterations, float p) {
    if (dummy) {
        return 0.0;
    }

    if (forceLowEstimate) {
//        std::cout << "Returning very low estimate from " << operationName << std::endl;
        return std::numeric_limits<float>::max();
    }

    if (!estimator_ptr->canEstimate()) {
//        std::cout << "Estimator for \"" << operationName << "\" not sufficiently trained yet, returning 0.0 as cost." << std::endl;
        // Not sure what should be returned here
        return 0.0;
    }

    double splineEstimate = 0.0;
    if (pDependent) {
        splineEstimate = estimator_ptr->dEstimate(input_size, n_iterations, p);
    }
    else if (inputSizeDependent) {
        splineEstimate = estimator_ptr->dEstimate(input_size, n_iterations);
    }
    else {
        splineEstimate = estimator_ptr->dEstimate(n_iterations);
    }
    return splineEstimate;
}

Operation &Operation::operator=(const Operation &other) {
    id = other.id;
    operationName = other.operationName;
    dummy = other.dummy;
    operationTimer = other.operationTimer;
    operation = other.operation;
    self = other.self;
    estimator_ptr = other.estimator_ptr;
    forceLowEstimate = other.forceLowEstimate;
    inputSizeDependent = other.inputSizeDependent;
    pDependent = other.pDependent;
    return *this;
}

Operation::Operation(const Operation &other) {
    *this = other;
}

void Operation::operator()(int input_size, int n_iterations, float p) {
    if (dummy) {
        throw Exception("Operation is a dummy, cannot perform operation!");
    }

    double estimate = cost(input_size, n_iterations, p);
    std::cout << "Starting " << operationName << "...";
    operationTimer.reset();
    operationTimer.start();
    (*self.*operation)();
    operationTimer.stop();

    double actual_cost = operationTimer.ms();

    std::cout << " Total time: " << actual_cost << "ms (estimated: " << estimate << "ms)" << std::endl;

    if (pDependent) {
        estimator_ptr->train(input_size, n_iterations, p, actual_cost);
    }
    else if (inputSizeDependent) {
        estimator_ptr->train(input_size, n_iterations, actual_cost);
    }
    else {
        estimator_ptr->train(n_iterations, actual_cost);
    }
}

void Operation::saveEstimator() {
//    std::cout << "Saving " << operationName << std::endl;
    if (estimator_ptr != nullptr) {
        estimator_ptr->save();
    }
    else if (!dummy) {
        std::cerr << "Operation (" << operationName << ") that is not a dummy does not have an estimator!" << std::endl;
    }
}
