#include <cmath>
#include <utility>
#include <utils.h>
#include <experimental/filesystem>


std::pair<int, int> getTwoFactors(int integer) {
    int first_factor = integer;
    int second_factor = 1;

    int square_root_floored = (int) sqrt(integer);

    for (int i = square_root_floored; i > 1; i--) {
        if (integer % i == 0) {
            first_factor = i;
            second_factor = integer / first_factor;
            break;
        }
    }

    return std::make_pair(first_factor, second_factor);
}

Timer::Timer() :
    stopped(true),
    start_time(time_point_t::min()),
    end_time(time_point_t::min()),
    total_time(duration_t::zero())
{
}

Timer::duration_t Timer::getTotalDuration() const {
    return stopped
           ? total_time
           : total_time + (now() - start_time);
}

Timer::time_point_t Timer::now() const {
    return my_clock_t::now();
}

double Timer::s() const {
    return ms() / 1000.0;
}

double Timer::ms() const {
    return us() / 1000.0;
}

double Timer::us() const {
    return ns() / 1000.0;
}

double Timer::ns() const {
    return (double) std::chrono::duration_cast<std::chrono::nanoseconds>(getTotalDuration()).count();
}

void Timer::start() {
    if (!stopped) {
        throw TimerException("Attempted to start an already running Timer!");
    }
    start_time = now();
    stopped = false;
}

void Timer::stop() {
    if (stopped) {
        throw TimerException("Attempted to stop an already stopped Timer!");
    }
    end_time = now();
    total_time += end_time - start_time;
    stopped = true;
}

void Timer::reset() {
    start_time = now();
    total_time = duration_t::zero();
}

void Timer::restart() {
    stop();
    reset();
    start();
}

Timer operator+(const Timer &lhs, const Timer &rhs) {
    Timer t;
    t.total_time = lhs.total_time + rhs.total_time;
    return t;
}

Timer operator-(const Timer &lhs, const Timer &rhs) {
    Timer t;
    t.total_time = lhs.total_time - rhs.total_time;
    return t;
}

bool file_exists(const std::string& filename) {
    std::ifstream f(filename.c_str());
    return f.good();
}

bool file_exists(const std::experimental::filesystem::path& filename) {
    return std::experimental::filesystem::exists(filename);
}

void _gpuAssert(cudaError_t code, const char *file, int line, bool abort)
{
    if (code != cudaSuccess)
    {
        fprintf(stderr,"GPUassert: %s %s %d\n", cudaGetErrorString(code), file, line);
        if (abort) {
            exit(code);
        }
    }
}

nlohmann::json json;
int cuda_device = 1;
